# coding: utf-8
"""
    MfTutorial.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import get_color
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from DraggablePoints import DraggablePoints
from SvdPrecision import plot_text

FONTSIZE = 200

def solve_efgh_equation():
    """
        Ｐ・Ｃ ＝ Ｍ ＝ Ｕ・Ｓ・Ｖ ＝ (Ｕ・Ｘ)・(Ｙ・Ｖ)

         Ｘ  ・ Ｙ  ＝   Ｓ

        [a b]  [e f] = [s1  0]
        [c d]  [g h]   [0  s2]

        (1) a*e + b*g = s1
        (2) a*f + b*h =  0
        (3) c*e + d*g =  0
        (4) c*f + d*h = s2
    """
    from sympy import var, solve
    a, b, c, d, e, f, g, h, s1, s2 = var('a b c d e f g h s1 s2')
    E1 = a*e + b*g - s1
    E2 = a*f + b*h
    E3 = c*e + d*g
    E4 = c*f + d*h - s2
    sols = solve([E1, E2, E3, E4], [e, f, g, h])
    print(sols)
    """
       {e: d*s1/(a*d - b*c), f: -b*s2/(a*d - b*c), g: -c*s1/(a*d - b*c), h: a*s2/(a*d - b*c)}

        e:  d*s1/(a*d - b*c)
        f: -b*s2/(a*d - b*c)
        g: -c*s1/(a*d - b*c)
        h:  a*s2/(a*d - b*c)
    """

def solve_efgh(a, b, c, d, s1, s2):
    det = a*d - b*c
    e = d*s1/det
    f = -b*s2/det
    g = -c*s1/det
    h = a*s2/det
    return e, f, g, h

def demo_func():
    fig = dplt.figure(figsize=(20,10))
    demo_func_impl(fig)
    dplt.show()

def generate_random_points():
    a = np.random.rand()*2 - 1
    b = np.random.rand()*2 - 1
    c = np.random.rand()*2 - 1
    d = np.random.rand()*2 - 1
    points = [(a,b), (c,d)]
    return points

def demo_func_impl(fig):
    q = np.linspace(0.01, 0.6, 600)
    rg_list = [50, 24]
    d_list = [1, 3]
    md = ModeledData(q, 300, rg_list=rg_list, d_list=d_list)
    if False:
        dplt.push()
        md.plot_components()
        dplt.pop()
    M = md.get_data()

    rank = len(rg_list)

    U, s, VT = np.linalg.svd(M)

    Us_ = np.dot( U[:,0:rank], np.diag( s[0:rank] ) )
    M_  = np.dot( Us_, VT[0:rank,:] )

    gs = GridSpec(2,4)
    axes = []
    for i in range(2):
        arow = []
        for k in range(4):
            proj = '3d' if k == 0 else None
            arow.append(fig.add_subplot(gs[i,k], projection=proj))
        axes.append(arow)
    ax4, ax1, ax2, ax3  = axes[0]

    for i in range(rank):
        ax1.plot(q, U[:,i])
        ax3.plot(VT[i,:])

    simple_plot_3d(ax4, M_, x=q)
    ns = 5
    ax2.plot(s[0:ns], ':', color='gray')
    for k in range(ns):
        ax2.plot(k, s[k], 'o', color=get_color(k))

    plot_text(ax1, "$U$", fontsize=FONTSIZE)
    plot_text(ax2, "$\Sigma$", fontsize=FONTSIZE)
    plot_text(ax3, "$V^T$", fontsize=FONTSIZE)

    ax8, ax5, ax6, ax7  = axes[1]
    points = generate_random_points()
    draw_PQ([ax5, ax7, ax8], q, points, U, s, VT)

    ax6.set_xlim(-2, 2)
    ax6.set_ylim(-2, 2)
    colors = ['red', 'yellow']

    objects = []
    for p, c_ in zip(points, colors):
        objects.append(ax6.plot(*p, 'o', color=c_))

    fig.tight_layout()
    return q, U, s, VT, axes, points, objects

def draw_PQ(axes, q, points, U, s, VT):
    ax5, ax7, ax8 = axes
    s1 = s[0]
    s2 = s[1]
    rank = len(points)
    a, b = points[0]
    c, d = points[1]
    e, f, g, h = solve_efgh(a, b, c, d, s1, s2)
    X = np.array([[a, b],[c, d]])
    Y = np.array([[e, f],[g, h]])
    print(np.dot(X, Y))
    print(s1, s2)
    P = np.dot(U[:, 0:rank], X)
    C = np.dot(Y, VT[0:rank,:])
    Q = np.dot(P, C)

    for i in range(rank):
        ax5.plot(q, P[:,i])
        ax7.plot(C[i,:])
    simple_plot_3d(ax8, Q, x=q)
    plot_text(ax5, "$P$", fontsize=FONTSIZE)
    plot_text(ax7, "$C$", fontsize=FONTSIZE)

class MfTutorialDemo(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.currently_dragging = False
        self.current_n = None
        self.polygon = None
        Dialog.__init__(self, parent, "MF Tutorial Demo", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(20,10))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_press)
        self.mpl_canvas.mpl_connect('button_release_event', self.on_release)
        self.mpl_canvas.mpl_connect('motion_notify_event', self.on_motion)

        self.q, self.U, self.s, self.VT, self.axes, points, self.objects = demo_func_impl(fig)
        self.ax6 = self.axes[1][2]
        self.points = points
        self.draw_points()
        self.mpl_canvas.draw()

    def draw_points(self):
        points = self.points
        points_ = np.array(points)
        #  print(points)
        # print(points_)
        for point, pobj in zip(points, self.objects):
            p, = pobj
            p.set_xdata([point[0]])
            p.set_ydata([point[1]])

        self.px = points_[:,0]
        self.py = points_[:,1]
        vx = np.sum(self.px)
        vy = np.sum(self.py)
        poly_points = [(0, 0), points[0], (vx,vy), points[1]]
        if self.polygon is None:
            self.polygon = Polygon(poly_points, alpha=0.1)
        else:
            self.polygon.set_xy(poly_points)

        self.ax6.add_patch(self.polygon)

    def on_press(self, event):
        if event.dblclick or True:
            self.draw_next_rand_point(event.button)
            return

        if event.inaxes != self.ax6:
            return

        print('on_press')
        self.currently_dragging = True
        dist = (self.px - event.xdata)**2 + (self.py - event.ydata)**2
        n = np.argmin(dist)
        print('n=', n)
        self.current_n = n

    def on_release(self, event):
        if not self.currently_dragging:
            return

        print('on_release')
        self.update_point(event)
        self.currently_dragging = False
        self.current_n = None
        self.redraw_decomp()
        self.mpl_canvas.draw()

    def on_motion(self, event):
        if not self.currently_dragging:
            return

        print('on_motion')
        n = self.current_n
        p, = self.objects[n]
        p.set_xdata([event.xdata])
        p.set_ydata([event.ydata])
        self.mpl_canvas.draw()

    def update_point(self, event):
        n = self.current_n
        self.points[n] = (event.xdata, event.ydata)
        self.draw_points()

    def redraw_decomp(self):
        ax8, ax5, _, ax7 = self.axes[1]
        for ax in [ax5, ax7, ax8]:
            ax.cla()
        draw_PQ([ax5, ax7, ax8], self.q, self.points, self.U, self.s, self.VT)

    def draw_next_rand_point(self, button):
        print('draw_next_rand_point')
        n = 0
        while True:
            # print([n])
            points = generate_random_points()
            if button == 1 or self.positive_PC(points):
                break
            n += 1
            if n > 100000:
                break
        print([n])
        self.points = points
        self.draw_points()
        self.redraw_decomp()
        self.mpl_canvas.draw()

    def positive_PC(self, points):
        rank = len(points)
        a, b = points[0]
        c, d = points[1]
        s1, s2 = self.s[0:2]
        e, f, g, h = solve_efgh(a, b, c, d, s1, s2)
        X = np.array([[a, b],[c, d]])
        Y = np.array([[e, f],[g, h]])
        # print(np.dot(X, Y))
        # print(s1, s2)
        P = np.dot(self.U[:, 0:rank], X)
        C = np.dot(Y, self.VT[0:rank,:])
        return np.all(P >= 0) and np.all(C >= 0)



import random
import numpy as np

import matplotlib
import matplotlib.pyplot as plt

import matplotlib.animation as animation


fps = 30
nSeconds = 5
snapshots = [ np.random.rand(5,5) for _ in range( nSeconds * fps ) ]

# First set up the figure, the axis, and the plot element we want to animate
fig = plt.figure( figsize=(8,8) )

a = snapshots[0]
im = plt.imshow(a, interpolation='none', aspect='auto', vmin=0, vmax=1)

def animate_func(i):
    if i % fps == 0:
        print( '.', end ='' )

    im.set_array(snapshots[i])
    return [im]

anim = animation.FuncAnimation(
                               fig, 
                               animate_func, 
                               frames = nSeconds * fps,
                               interval = 1000 / fps, # in ms
                               )

anim.save('test_anim.mp4', fps=fps, extra_args=['-vcodec', 'libx264'])

print('Done!')

plt.show()



# coding: utf-8
"""
    MethodFileSimulator.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from MethodFile import MethodFile

"""
 0  Pressurization
 1  FlowModulation
 2  Sample Filling (pumpB)
 3  FlowModulation
 4  Sample Filling (pumpA')
 5  FlowModulation
 6  Sample Filling (pumpC)
 7  FlowModulation
 8  FlowModulation
 9  u-fluidics wash fast flow II
10  Total flow rate modulation
11  100% pump A' (buffer)
12  pumpA' -> pumpA&C
13  pump A&C (Base Sample)
14  Y channel initialize
15  const flow rate (pump A, B, C)
16  FlowModulation
17  Waste
18  FlowModulation
19  const flow rate (pump A, B)
20  Linearly changing flow rate
21  const. uL/min (pump B)
22  FlowModulation
23  waste
24  FlowModulation
25  Pressurization & Waiting
"""

TOTAL_RATE_COLUMNS = [2, 4, 6, 10]
RATE_B_START_COLUMN = 4
RATE_B_STOP_COLUMN = 5

def set_default_params(self):
    self.delay_factor = 8
    if self.eno_interval is None:
        self.eno_interval = 1
    self.time_delta = self.eno_interval/2

class MethodFileSimulator(MethodFile):
    def __init__(self, filepath):
        MethodFile.__init__(self, filepath)
        self.eno_interval = None
        set_default_params(self)

    def set_params(self, delay_factor=None, default_init=False, debug=False):
        if default_init:
            # this call should not be a method call to allow uses without inheritance.
            set_default_params(self)

        if delay_factor is None:
            delay_factor = self.delay_factor

        i_, j_, k_ = self.guess_critical_points()
        p_ = i_ - 4     # temp
        j_1 = j_ - 1

        integ_duration = 0
        for k, values in enumerate(self.rows):
            assert values[0] == k + 1

            duration = values[1]/60
            integ_duration += duration
            if debug:
                print([k], values + [integ_duration])
            if k == j_1:    # const flow rate (pump A, B)
                self.start_duration = integ_duration
            elif k == p_:       # Total flow rate modulation
                self.pre_start_time = integ_duration
            elif k == i_:       # name == 'Y channel initialize
                frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.y_channel_time = integ_duration + delay_factor/frate_total
            elif k == j_:       # Linearly changing flow rate
                self.stop_duration = integ_duration
                self.start_frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.start_time = self.start_duration + delay_factor/self.start_frate_total
                self.start_conc = values[RATE_B_START_COLUMN]/self.start_frate_total
                rate_b_stop = values[RATE_B_STOP_COLUMN]
            elif k == k_:       # const. uL/min (pump B)
                self.stop_frate_total = np.sum([values[k] for k in TOTAL_RATE_COLUMNS])
                self.stop_time = self.stop_duration + delay_factor/self.stop_frate_total
                self.stop_conc = rate_b_stop/self.stop_frate_total
            last_values = values

        self.end_time = integ_duration + delay_factor/np.sum([last_values[k] for k in TOTAL_RATE_COLUMNS])

        if debug:
            for r in [  self.start_duration,
                        self.pre_start_time,
                        self.y_channel_time,
                        self.start_frate_total,
                        self.stop_duration,
                        self.stop_frate_total,
                        self.start_time,
                        self.stop_time,
                        self.start_conc,
                        self.stop_conc,
                        ]:
                print(r)



"""

    Test.TestAllData.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
from molass_legacy._MOLASS.SerialSettings import set_setting, clear_temporary_settings
from DataUtils import get_pytools_folder, serial_folder_walk
from molass_legacy.Batch.StandardProcedure import StandardProcedure

counter = 0
def test_all_data(callback_all, analysis_copy=False, logger=None, restart_folder=None, index_csv_fh=None):
    set_setting("test_pattern", 0)

    if restart_folder is not None:
        restart_folder = restart_folder.replace("\\", "/")

    sp = StandardProcedure()

    def walk_callback(in_folder, uv_folder=None, plot=False):
        global counter
        if restart_folder is not None:
            if in_folder < restart_folder:
                counter += 1
                print("skipping", in_folder)
                return True, None

        clear_temporary_settings()
        if analysis_copy:
            from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
            sd_ = sp.load_old_way(in_folder)
            try:
                pre_recog = PreliminaryRecognition(sd_)
                sd = sd_._get_analysis_copy_impl(pre_recog)
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(logger, "failed to get an analysis_copy: ", n=10)
                sd = sd_
        else:
            sd = sp.load_old_way(in_folder)

        callback_all(in_folder, sd, counter, index_csv_fh)
        counter += 1
        return True, None

    data_folder = get_pytools_folder() + '/Data'
    serial_folder_walk(data_folder, walk_callback)



"""
    Simulative.SeveralStudies.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

def several_studies_impl(lrf_src, guess_info, parent):
    print("several_studies_impl")
    ret_params = lrf_src.draw()     # this is currently equivalent to onthfly_test in StcEstimator.py
    print("ret_params =", ret_params)


"""
    DraggableLine.py

    borrowed, adapted and renamed from
        Interactive Line in matplotlib
        http://stackoverflow.com/questions/34855074/interactive-line-in-matplotlib
"""
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import numpy as np

class DraggableLine(object):
    def __init__(self, line, dual=None, epsilon=1.0, select_cb=None, update_cb=None, vertical_only=False,
                    sync_widgets=None ):
        canvas = line.figure.canvas
        self.canvas = canvas
        self.line = line
        self.dual = dual
        self.epsilon = epsilon
        self.select_cb = select_cb
        self.update_cb = update_cb
        self.vertical_only  = vertical_only
        self.sync_widgets   = sync_widgets
        self.axes = line.axes
        self.xs = list(line.get_xdata())
        self.ys = list(line.get_ydata())

        self.ind = None
        self.has_moved = False

        canvas.mpl_connect('button_press_event', self.button_press_callback)
        canvas.mpl_connect('button_release_event', self.button_release_callback)
        canvas.mpl_connect('motion_notify_event', self.motion_notify_callback)

    def get_ind(self, event):
        if event.inaxes !=self.axes or event.xdata is None or event.ydata is None:
            return None

        x = np.array(self.line.get_xdata())
        y = np.array(self.line.get_ydata())
        d = np.sqrt((x-event.xdata)**2 + (y - event.ydata)**2)
        if min(d) > self.epsilon:
            return None
        if d[0] < d[1]:
            return 0
        else:
            return 1

    def button_press_callback(self, event, level=0 ):
        if event.button != 1:
            return
        self.ind = self.get_ind(event)
        if self.ind is None: return

        print( 'button_press_callback: ind=', self.ind, 'level=', level )

        if self.select_cb is not None:
            self.select_cb( self.ind )

        self.line.set_animated(True)
        if self.dual is not None:
            self.dual.set( markevery=[ self.ind ] )
            self.dual.set_animated(True)

        self.canvas.draw()
        self.background = self.canvas.copy_from_bbox(self.line.axes.bbox)

        self.axes.draw_artist(self.line)
        self.canvas.blit(self.axes.bbox)
        self.has_moved = False
        if level == 0:
            print( 'len(sync_widgets)=', len(self.sync_widgets) )
            for w in self.sync_widgets:
                if w != self:
                    w.button_press_callback( event, level=1 )

    def button_release_callback(self, event):
        if event.button != 1:
            return
        self.ind = None
        self.line.set_animated(False)
        if self.dual is not None:
            self.dual.set_animated(False)
        self.background = None
        if self.update_cb is not None and self.has_moved:
            self.update_cb( self )
        self.line.figure.canvas.draw()

    def motion_notify_callback(self, event):
        if event.inaxes != self.line.axes:
            return
        if event.button != 1:
            return
        if self.ind is None:
            return

        if not self.vertical_only:
            self.xs[self.ind] = int( event.xdata + 0.5 )
        self.ys[self.ind] = event.ydata
        self.line.set_data(self.xs, self.ys)
        if self.dual is not None:
            self.dual.set_data(self.xs, self.ys)

        self.canvas.restore_region(self.background)
        self.axes.draw_artist(self.line)
        if self.dual is not None:
            self.axes.draw_artist(self.dual)
        self.canvas.blit(self.axes.bbox)
        self.has_moved = True

    def unselect( self ):
        if self.dual is not None:
            self.dual.set( markevery=[] )

if __name__ == '__main__':

    fig, ax = plt.subplots()
    line1 = Line2D([0, 1], [0, 1], marker = 'o', markersize=7, markerfacecolor = 'pink')
    line2 = Line2D([0, 1], [0, 1], linestyle='None', marker = 'o', markersize=7, markerfacecolor = 'red', markevery=[] )
    ax.add_line(line1)
    ax.add_line(line2)

    def select_cb( ind ):
        print( 'select_cb: ind=', ind )

    def update_cb( obj ):
        print( 'update_cb' )

    linebuilder = DraggableLine(line1, dual=line2, select_cb=select_cb, update_cb=update_cb )
    # linebuilder = DraggableLine(line1 )

    ax.set_title('click to create lines')
    ax.set_xlim(-2,2)
    ax.set_ylim(-2,2)
    plt.show()


"""

    OurImporter.py

    Copyright (c) 2024, Masatsuyo Takahashi, KEK-PF

"""
import sys
import importlib.util

"""
    https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
    https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
"""
def import_module_from_path(module_name, path):
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module



# coding: utf-8
"""
    Scattering3dPlot.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
import os
import copy
import numpy                as np
from bisect                 import bisect_right
from mpl_toolkits.mplot3d   import Axes3D, proj3d
from molass_legacy.KekLib.BasicUtils             import get_caller_module, get_filename_extension
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.TkUtils                import is_low_resolution, split_geometry
from CanvasFrame            import CanvasFrame
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting, get_xray_picking
from ScatteringBaseCorrector    import ScatteringBaseCorrector
from molass_legacy.KekLib.TkCustomWidgets        import FolderEntry
from DevSettings            import get_dev_setting, set_dev_setting, ITEM_DEFAULTS as DEV_ITEM_DEFAULTS
from ScatteringViewUtils    import compute_baselines, draw_3d_scattering, ZLIM_EXPAND

class Scattering3dPlotFrame( Tk.Frame ):
    def __init__( self, parent, dialog, serial_data, mapper, data, title, i, sim_base=None ):
        # experimental = i == 1
        experimental = False
        self.plot_no        = i
        self.parent         = parent
        self.dialog         = dialog
        self.sd             = serial_data
        self.qvector        = serial_data.qvector
        self.jvector        = serial_data.jvector
        self.index          = bisect_right( self.qvector, get_xray_picking() )
        self.xray_curve_y   = serial_data.xray_curve.y
        self.xray_slice     = serial_data.xray_slice
        self.data           = data
        self.data_list      = [ data ]
        self.title          = title
        self.sim_base       = sim_base
        self.experimental   = experimental
        self.el_x   = np.ones( len(self.jvector) ) * self.qvector[self.index]
        self.el_y   = self.jvector
        self.el_z   = self.xray_curve_y
        self.suppress_on_draw   = False

        opt_params  = mapper.opt_params

        self.baseline_pair_list = []

        data_copy = copy.deepcopy( data )
        corrector = ScatteringBaseCorrector(
                                serial_data.jvector,
                                serial_data.qvector,
                                data_copy,
                                curve=serial_data.xray_curve,
                                affine_info=mapper.get_affine_info(),
                                inty_curve_y=self.xray_curve_y,
                                baseline_opt=opt_params['xray_baseline_opt'],
                                baseline_degree=opt_params['xray_baseline_const_opt'] + 1,
                                need_adjustment=opt_params['xray_baseline_adjust'] == 1,
                                parent=self )
        self.baseline_pair_list.append( compute_baselines( self.qvector, self.index, corrector ) )
        self.corrected_data = data_copy

        Tk.Frame.__init__( self, parent )

        self.body( self )
        self.buttonbox()
        self.draw()
        self.create_popup_menu()

    def body( self, body_frame ):   # overrides parent class method

        base_frame = Tk.Frame( body_frame );
        # base_frame.pack( expand=1, fill=Tk.BOTH, padx=20, pady=10 )
        base_frame.pack( expand=1, fill=Tk.BOTH )

        figsize = ( 8, 8 ) if is_low_resolution() else ( 10, 10 )

        self.canvas_frame = canvas_frame = CanvasFrame( base_frame, figsize=figsize )
        canvas_frame.pack()
        fig = self.canvas_frame.fig
        ax1 = fig.add_subplot( 111, projection='3d' )
        self.axes = [ ax1 ]

        # fig.tight_layout()
        self.canvas_frame.fig.canvas.mpl_connect( 'draw_event', self.on_draw )
        self.canvas_frame.fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        frame = self

        box1 = Tk.Frame( frame )
        box1.pack( fill=Tk.X, anchor=Tk.CENTER )
        self.button_frame = Tk.Frame( box1 )
        self.button_frame.grid( row=0, column=0 )
        box1.columnconfigure( 0, weight=1 )
        # box1.columnconfigure( 1, weight=1 )

        self.cunstom_button_box_add( self.button_frame )

    def cunstom_button_box_add( self, frame ):
        self.vsa_only = Tk.IntVar()
        self.vsa_only.set( 1 )
        self.vsa_only_cb = Tk.Checkbutton( frame, text="Very Small Angle only", variable=self.vsa_only )
        self.vsa_only_cb.grid( row=0, column=0, padx=5 )

        space = Tk.Frame( frame, width=10 )
        space.grid( row=0, column=1, padx=10 )

        label = Tk.Label( frame, text="Zlim expand ratio:" )
        label.grid( row=0, column=2 )
        self.zlim_expand = Tk.DoubleVar()
        self.zlim_expand.set( ZLIM_EXPAND )
        entry = Tk.Entry( frame, textvariable=self.zlim_expand, width=6, justify=Tk.RIGHT )
        entry.grid( row=0, column=3 )

        space = Tk.Frame( frame, width=10 )
        space.grid( row=0, column=6, padx=10 )

        self.redraw_btn = Tk.Button( frame, text="Redraw", command=self.redraw )
        self.redraw_btn.grid( row=0, column=4, padx=10 )

        view_init = Tk.Button( frame, text="View Init", command=self.view_init )
        view_init.grid( row=0, column=5, padx=10 )

        if False:
            show_save = Tk.Button( frame, text="Save Dialog", command=self.show_save_dialog )
            show_save.grid( row=0, column=6, padx=10 )

    def create_popup_menu( self ):
        self.popup_menu = Tk.Menu( self, tearoff=0 )
        self.popup_menu.add_command( label='Guinier analysis animation', command=lambda: self.guinier_analysis() )
        self.popup_menu.add_command( label='Guinier analysis animation Corrected', command=lambda: self.guinier_analysis(corrected=True) )
        self.popup_menu.add_command( label='Rg comparison (GuinierDiffViewer)', command=self.show_guinier_viewer )

        # self.draw()
    def draw( self ):
        try:
            zlim_expand = self.zlim_expand.get()
        except:
            zlim_expand = None

        n = 0
        ax  = self.axes[n]
        data    = self.data_list[n]
        vsa_base_list, all_base_list    = self.baseline_pair_list[n]
        very_small_angle_only = self.vsa_only.get() == 1

        draw_3d_scattering( ax, data, self.qvector, self.index, self.xray_curve_y,
                            self.title,
                            vsa_base_list, all_base_list,
                            zlim_expand=zlim_expand, experimental=self.experimental,
                            very_small_angle_only=very_small_angle_only,
                            sim_base=self.sim_base,
                            )

        self.canvas_frame.show()

    def redraw( self ):
        for ax in self.axes:
            ax.cla()
        self.draw()
        self.canvas_frame.show()

    def view_init( self ):
        for ax in self.axes:
            ax.view_init()
        self.canvas_frame.show()

    def on_draw( self, event ):
        # print( 'on_draw' )
        if self.dialog.sync_view.get() == 0:
            return

        if self.suppress_on_draw:
            return

        this_ax = self.axes[0]
        other_frame = self.dialog.plot_frames[1-self.plot_no]
        # https://stackoverflow.com/questions/47610614/get-viewing-camera-angles-in-matplotlib-3d-plot
        other_frame.suppress_on_draw = True

        zlim_expand_ = self.zlim_expand.get()
        if other_frame.zlim_expand.get() != zlim_expand_:
            other_frame.zlim_expand.set(zlim_expand_)
            other_frame.redraw()

        other_frame.axes[0].view_init( this_ax.elev, this_ax.azim )
        other_frame.canvas_frame.show()
        other_frame.update()

        other_frame.suppress_on_draw = False

    def compute_projected_xy( self ):
        ax = self.axes[0]
        # print( ax.get_proj() )
        projection = ax.get_proj()
        x, y, _ = proj3d.proj_transform( self.el_x, self.el_y, self.el_z, projection )
        xy = np.vstack( [x, y] ).T
        ret = ax.transData.transform( xy )
        print( 'ret.shape=', ret.shape )
        self.x_scr = ret[:,0]
        self.y_scr = ret[:,1]
        if False:
            for i in range(5):
                print( 'point[%d]=' % i, ( self.x[i], self.y[i], self.r[i] ) )
                # print( 'transformed point[%d]=' % i, (x[i], y[i]) )
                print( 'transformed screen point[%d]=' % i, (self.x_scr[i], self.y_scr[i]) )

    def on_button_press( self, event ):
        if event.button == 3:
            self.on_right_button_press( event )
            return

    def on_right_button_press( self, event ):
        print( 'on_right_button_press' )

        if event.inaxes not in self.axes:
            return

        self.compute_projected_xy()

        print( 'x=', event.x )
        print( 'y=', event.y )
        w, h, x, y = split_geometry( self.canvas_frame.mpl_canvas_widget.winfo_geometry() )
        h_ = h
        print( 'self.mpl_canvas_widget.winfo_geometry()=', (w, h, x, y) )

        w, h, x, y = split_geometry( self.dialog.winfo_geometry() )
        # print( 'self.winfo_geometry()=', (w, h, x, y) )
        # TODO: better way to get the mouse cursor position

        elution_no = self.guess_pointed_elution_no( event )
        if elution_no is None:
            return

        self.elution_no = elution_no
        x_  = x if self.plot_no == 0 else x + w//2

        # self.folder_no.set( folder_no )
        self.popup_menu.entryconfig( 0, label="Guinier analysis animation on %d-th elution" % elution_no )
        self.popup_menu.entryconfig( 1, label="Guinier analysis animation on %d-th elution corrected" % elution_no )
        self.popup_menu.post( int( x_ + event.x + 10 ), int( y + h_ - event.y + 50 ) )

    def guess_pointed_elution_no( self, event ):
        dist = ( self.x_scr - event.x )**2 + ( self.y_scr - event.y )**2
        i = np.argmin( dist )
        return i

    def update_cursor( self, cursor ):
        for w in [self.canvas_frame.mpl_canvas_widget]:
            w.config( cursor=cursor )

    def guinier_analysis( self, elution_no=None, corrected=False ):
        if elution_no is None:
            elution_no = self.elution_no

        print( 'guinier_analysis: ', elution_no )
        self.update_cursor( 'wait' )

        from SimpleGuinier          import SimpleGuinier
        from molass_legacy.AutorgKekAdapter       import AutorgKekAdapter
        from SimpleGuinierAnimation import SimpleGuinierAnimation
        from CanvasDialog           import CanvasDialog

        if corrected:
            data    = self.corrected_data
        else:
            data    = self.data

        s_data  = data[elution_no,]
        guinier = SimpleGuinier( s_data, anim_data=True )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()

        dialog = CanvasDialog( "SimpleGuinier Animation", parent=self )
        anim_iter_max = 300
        title = 'Guinier analysis of the ' + str(elution_no) + '-th elution'

        def anim_func( fig  ):
            anim = SimpleGuinierAnimation( guinier, title=title, result_rg=adapter_result.Rg  )
            anim.draw( fig, anim_iter_max=anim_iter_max )
            dialog.mpl_canvas.draw()

        self.update_cursor( '' )

        dialog.show( anim_func, figsize=( 18, 5 ) )

    def show_guinier_viewer( self ):
        self.config( cursor='wait' )
        self.update()

        from ScatteringBaseUtil     import apply_baseline_correction_impl
        from GuinierDiffViewer      import GuinierDiffViewer

        sd  = self.sd

        usable_slice = sd.get_usable_slice()
        mapper = self.dialog.mapper
        # data1 = self.dialog.data_list[0]
        data1 = self.data
        data2 = copy.deepcopy( data1 )

        mapped_info = mapper.get_mapped_info()
        apply_baseline_correction_impl(
                sd.jvector, sd.qvector, data2,
                mapped_info,
                ecuve=sd.xray_curve,
                )

        self.config( cursor='' )
        viewer = GuinierDiffViewer( self.dialog, sd.qvector, sd.ivector, data1, data2, self.elution_no, usable_slice=usable_slice )
        self.dialog.update()

    def show_save_dialog( self ):
        dialog = ScatteringSaveDialog( self.dialog )
        dialog.show()

class ScatteringSaveDialog( Dialog ):
    def __init__( self, parent ):
        self.parent = parent
        self.caller_module = get_caller_module( level=2 )

    def show( self ):
        title = "Absorbance Data Save"
        Dialog.__init__( self, self.parent, title, auto_geometry=False, geometry_cb=self.adjust_geometry )

    def adjust_geometry( self ):
        w, h, x, y = split_geometry( self.parent.geometry() )
        self.geometry("+%d+%d" % (self.parent.winfo_rootx() + w//2,
                                  self.parent.winfo_rooty() + int(h*0.7) ))

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        label_frame = Tk.Frame(body_frame)
        label_frame.pack( padx=50, pady=10 )

        guide = Tk.Label(label_frame, text='Select what and where (file name) to save and press "OK"')
        guide.pack()

        detail_frame = Tk.Frame(body_frame)
        detail_frame.pack( padx=50, pady=10 )

        grid_row = 0
        self.xb_save = Tk.IntVar()

        cb = Tk.Checkbutton( detail_frame, text="enable X-ray Scattering Baseline Save",
                                variable=self.xb_save  )
        cb.grid( row=grid_row, column=0, sticky=Tk.W   )

        self.xb_folder = Tk.StringVar()
        folder_path = os.path.join( get_setting( 'analysis_folder' ), 'xray_scattering_base' ).replace( '\\', '/' )
        self.xb_folder.set( folder_path )
        folder_entry = FolderEntry( detail_frame, textvariable=self.xb_folder, width=70,
                                            on_entry_cb=self.on_entry_xb_folder )
        folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        grid_row += 1
        self.filename_example   = os.path.split( self.parent.sd.datafiles[0] )[-1]
        self.filename_extention = '.' + get_filename_extension( self.filename_example )

        postfix_frame = Tk.Frame( detail_frame )
        postfix_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        postfix_label = Tk.Label( postfix_frame, text="filename postfix" )
        postfix_label.grid( row=0, column=0 )
        self.base_file_postfix = Tk.StringVar()
        self.base_file_postfix.set( "_base" )

        self.base_file_postfix_entry = Tk.Entry( postfix_frame, textvariable=self.base_file_postfix, width=10 )
        self.base_file_postfix_entry.grid( row=0, column=1, padx=5 )

        as_in_label = Tk.Label( postfix_frame, text="as in" )
        as_in_label.grid( row=0, column=2 )

        self.postfix_eg = Tk.StringVar()
        self.postfix_eg_update()
        eg_label = Tk.Label( postfix_frame, textvariable=self.postfix_eg )
        eg_label.grid( row=0, column=3, padx=3 )

    def on_entry_xb_folder( self ):
        pass

    def postfix_eg_update( self ):
        filename = self.filename_example.replace( self.filename_extention, self.base_file_postfix.get() + self.filename_extention )
        self.postfix_eg.set( filename )

    def validate( self ):
        if self.xb_save.get() == 1:
            self.save_xray_scattering_base()
        return 1

    def save_xray_scattering_base( self ):
        for file in self.parent.sd.datafiles:
            d, f = os.path.split( file )
            print( f )

class Scattering3dPlotDialog( Dialog ):
    def __init__( self, parent, serial_data, mapper, data_list=None, title_list=None, sim_base=None ):
        self.parent = parent
        self.sd     = serial_data
        self.mapper = mapper
        if data_list is None:
            data_list   = [ serial_data.intensity_array ]
        self.data_list  = data_list
        if title_list is None:
            title_list  = []

            title = 'Xray Scattering from ' + get_setting( 'in_folder' )
            for n in range( len(self.data_list) ):
                sim_title = '' if n == 0 else 'Residual of '
                title_list.append( sim_title + title )

        self.title_list = title_list
        self.sim_base   = sim_base
        self.applied    = False

    def show( self ):
        title = "ScatteringThreeDimPlot"
        Dialog.__init__(self, self.parent, title )

    def body( self, body_frame ):
        tk_set_icon_portable( self )

        plot_frame = Tk.Frame( body_frame )
        plot_frame.pack()

        self.sync_view = Tk.IntVar()
        self.sync_view.set(0)

        self.plot_frames = []

        for i, data in enumerate(self.data_list):
            # add ", sim_base=self.sim_base" when plotting simulated base
            # plot3d_frame = Scattering3dPlotFrame( plot_frame, self, self.sd, self.mapper, data, self.title_list[i], i, sim_base=self.sim_base )
            plot3d_frame = Scattering3dPlotFrame( plot_frame, self, self.sd, self.mapper, data, self.title_list[i], i )
            plot3d_frame.grid( row=0, column=i )
            self.plot_frames.append( plot3d_frame )

        if len(self.plot_frames) > 1:
            self.sync_view.set(1)

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        frame = self

        box2 = Tk.Frame( frame )
        box2.pack( fill=Tk.X, expand=1 )

        frames = []
        for k in range(3):
            f = Tk.Frame( box2 )
            f.pack( side=Tk.LEFT, fill=Tk.X, expand=1 )
            frames.append( f )

        w = Tk.Button(frames[2], text="Cancel", width=10, command=self.cancel, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=10, pady=10)
        self.cancel_button = w

        w = Tk.Button(frames[2], text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=10, pady=10)
        self.ok_button = w

        self.apply_fitted_base = Tk.IntVar()
        self.apply_fitted_base.set(0)

        if len(self.plot_frames) > 1:
            w = Tk.Checkbutton(frames[2], text="Apply Fitted Correction", variable=self.apply_fitted_base )
            w.pack(side=Tk.RIGHT, padx=10, pady=10)

            cb = Tk.Checkbutton( frames[1], text="Synchronize view", variable=self.sync_view  )
            cb.pack( side=Tk.RIGHT, padx=20 )

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def apply( self ):  # overrides parent class method
        self.applied        = True



"""

    GuiMctTools.py

    Copyright (c) 2018-2022, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter import Tk

class GuiMctToolsMenu(Tk.Menu):
    def __init__(self, parent, dialog, menubar ):
        self.parent = parent
        self.dialog = dialog

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="MCT Tools", menu=self )
        self.add_command( label="Average Subtractor", command=self.show_average_subtractor_dialog )
        self.add_command( label="Individual Subtractor", command=self.show_individual_subtractor_dialog )
        self.add_command( label="Conc Normalizer", command=self.show_conc_normalizer_dialog )
        self.add_command( label="Elution Curve Picker", command=self.show_elution_curve_picker, state=Tk.DISABLED )
        self.add_command( label="Average Maker", command=self.show_average_maker)

    def update_states(self):
        state = Tk.NORMAL if self.dialog.dataset_is_ready else Tk.DISABLED
        for k in [3]:
            self.entryconfig(k, state=state)

    def show_average_subtractor_dialog(self):
        print('show_average_subtractor_dialog')
        from Microfluidics.AverageSubtractorDialog import AverageSubtractorDialog
        dialog = AverageSubtractorDialog(self.parent)
        dialog.show()

    def show_individual_subtractor_dialog(self):
        print('show_individual_subtractor_dialog')
        from Microfluidics.IndividualSubtractorDialog import IndividualSubtractorDialog
        dialog = IndividualSubtractorDialog(self.parent)
        dialog.show()

    def show_conc_normalizer_dialog(self):
        print('show_conc_normalizer_dialog')
        from Microfluidics.ConcNormalizerDialog import ConcNormalizerDialog
        dialog = ConcNormalizerDialog(self.parent)
        dialog.show()

    def show_elution_curve_picker(self):
        print('show_individual_subtractor_dialog')
        from molass_legacy.Tools.ElutionCurvePicker import ElutionCurvePicker
        parent = self.parent
        dialog = ElutionCurvePicker(parent, parent.pre_recog)
        dialog.show()
        if dialog.applied:
            parent.fig_frame.update_elution_curve()

    def show_average_maker(self):
        from SecTools.AverageMakerDialog import AverageMakerDialog
        dialog = AverageMakerDialog(self.parent)
        dialog.show()



# coding: utf-8
"""
    Boundary.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
from bisect import bisect_right
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data
from DataUtils import get_in_folder

WIDTH = 40
BOUNDARY_RATIO = 0.8

def get_boundary(qv, M, max_q):
    qlim_i = bisect_right(qv, max_q)

    sv_ratios = []
    if qlim_i+WIDTH > len(qv):
        qlim_i = len(qv) - WIDTH

    # print(qlim_i, qv[qlim_i])

    for i in range(0,qlim_i):
        s = np.linalg.svd(M[i:i+WIDTH,:])[1]
        sv_ratios.append(s[1]/s[0])

    sv_ratios = np.array(sv_ratios)
    fwd_ratios = sv_ratios[WIDTH:]/sv_ratios[0:-WIDTH]
    k = int(len(fwd_ratios)*BOUNDARY_RATIO)
    p = np.argpartition(fwd_ratios, k)
    m = np.min(p[k:])
    return m+WIDTH, fwd_ratios, qlim_i, sv_ratios, p[k:]

def demo(sd):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[0]
    f = range_[0]
    p = range_[1]
    t = range_[2]+1
    eslice = slice(f,t)

    m, fwd_ratios, qlim_i, sv_ratios, pk_ = get_boundary(qv, M[:,eslice], 0.3)
    aslice = slice(0,qlim_i)

    x = ecurve.x
    y = ecurve.y
    c = y[f:t].copy()
    c /= np.max(c)

    M_ = M[aslice,eslice]
    M1 = get_denoised_data(M_, rank=1)
    C1 = np.array([c])
    P1 = M1 @ np.linalg.pinv(C1)

    M2 = get_denoised_data(M_, rank=2)
    C2 = np.array([c, c**2])
    P2 = M2 @ np.linalg.pinv(C2)

    y1 = P1[:,0]
    y2 = P2[:,0]
    bq = P2[:,1]

    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21, 7))
    ax2.set_yscale('log')
    # ax3.set_yscale('log')

    fig.suptitle("Rank Boundary Determination for " + get_in_folder(), fontsize=20)
    ax1.set_title("Elution Curve and Analysis Range", fontsize=16)
    ax2.set_title("Scattering Curves", fontsize=16)
    ax3.set_title("S1/S0 Ratio Variation and B(q)", fontsize=16)

    ax1.set_xlabel('Eno')
    ax2.set_xlabel('Q')
    ax3.set_xlabel('Q')

    ax1.plot(x, y, color='orange')

    ymin, ymax = ax1.get_ylim()
    patch = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )

    ax1.add_patch(patch)

    qv_ = qv[aslice]
    ax2.plot(qv_, M[aslice,p], alpha=0.3, label='peak top')
    ax2.plot(qv_, y1, alpha=0.3, label='rank(1,1)')
    ax2.plot(qv_, y2, alpha=0.3, label='rank(2,2)')

    dw = (qv[WIDTH]-qv[0])      # assuming regular intervals
    ax3.plot(qv[0:qlim_i]+dw, sv_ratios, label='$bs_1/bs_0$')
    ax3.plot(qv[0:qlim_i-WIDTH]+dw, fwd_ratios, label='$(fs_1/fs_0)/(bs_1/bs_0)$')
    ax3.plot(qv[pk_]+dw, fwd_ratios[pk_], 'o', color='yellow', markersize=3, label='over %d %% points' % (100*BOUNDARY_RATIO))

    bx = qv[m]
    for ax in [ax2, ax3]:
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        ax.plot([bx, bx], [ymin, ymax], ':', color='red', label='rank boundary')

    axt = ax3.twinx()
    axt.plot(qv_, bq, color='pink', label='B(q)')
    axt.grid(False)

    ax3.set_ylabel('Ratio')
    axt.set_ylabel('B(q) Value')

    ax2.legend()
    ax3.legend(loc='center right')
    axt.legend(loc='upper right')
    fig.tight_layout()
    fig.subplots_adjust(top=0.88)

    plt.show()

def boundary_demo(sd):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[0]
    f = range_[0]
    p = range_[1]
    t = range_[2]+1
    eslice = slice(f,t)

    _, fwd_ratios, qlim_i, sv_ratios, pk_ = get_boundary(qv, M[:,eslice], 0.3)

    n = np.argmax(fwd_ratios)
    m = np.where(fwd_ratios[n:] < 1)[0][0] + n + 40

    fig, ax = plt.subplots()
    ax.set_yscale('log')

    y = M[:,p]
    ax.plot(qv, y)

    ymin, ymax = ax.get_ylim()

    for rank, fq, tq, color in [(2, qv[m-40], qv[m], 'pink'), (1, qv[m], qv[m+40], 'cyan')]:
        patch = Rectangle(
                (fq, ymin),     # (x,y)
                tq - fq,        # width
                ymax - ymin,    # height
                facecolor   = color,
                alpha       = 0.2,
                label = 'Rank %d' % rank,
            )

        ax.add_patch(patch)

    ax.legend()
    fig.tight_layout()
    plt.show()



"""

    GuiTutorials.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter     import Tk, is_empty_val
from molass_legacy._MOLASS.SerialSettings import get_setting
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox

CONJUGATE_GRADIENT = 5

class GuiTutorialsMenu(Tk.Menu):
    def __init__(self, parent, menubar ):
        self.parent = parent

        Tk.Menu.__init__(self, menubar, tearoff=0 )
        menubar.add_cascade( label="Tutorials", menu=self )
        self.add_command( label="Linear Transformation", command=self.show_lintran_tutorial )
        self.add_command( label="Matrix Inverse", command=self.show_matrix_inverse )
        self.add_command( label="Moore-Penrose Inverse", command=self.show_moore_penrose_inverse )
        self.add_command( label="SVD Animation", command=self.show_svd_animation )
        self.add_command( label="SVD (2D)", command=self.show_svd_tutorial_2d )
        self.add_command( label="SVD (3D)", command=self.show_svd_tutorial_3d )
        self.add_command( label="Matrix Factorization", command=self.show_mf_tutorial )
        # update CONJUGATE_GRADIENT above if the order No. of this entry changes.
        self.add_command( label="Conjugate Gradient Algorithm", command=self.show_conjugate_gradient, state=Tk.DISABLED )
        self.add_command( label="Affine Transformation of Baseline Adjustment", command=self.show_affine_transformation )
        self.add_command( label="Extrapolation with Spherical Particles", command=self.show_extrapolation_simulation )
        self.add_command( label="Noise and Sigma-width Dependency of Base Percentile Offset", command=self.show_bpo_dependency_demo )

    def update_states(self):
        state = Tk.NORMAL if self.parent.dataset_is_ready else Tk.DISABLED
        self.entryconfig(CONJUGATE_GRADIENT, state=state)

    def show_extrapolation_simulation( self ):
        from SimulatedExtrapolation import SimulatedExtrapolationDialog
        dialog = SimulatedExtrapolationDialog( 'Simulation', parent=self.parent )
        dialog.show()

    def show_bpo_dependency_demo(self):
        from BpoDependencyDemo import BpoDependencyDemo
        self.parent.config(cursor='wait')
        self.parent.update()
        dialog = BpoDependencyDemo(self.parent)
        self.parent.config(cursor='')
        dialog.show()

    def show_lintran_tutorial(self):
        latex_ok = self.parent.env_info.get_latex_availability()
        if False and latex_ok:
            ret = MessageBox.askokcancel("Warning",
                    "The application currently terminates\n"
                    + "after closing this tutorial\n"
                    + "due to a bug or misuse of some codes.\n"
                    + "Do you proceed anyway?",
                    parent=self.parent)
            if not ret:
                return

            from Tutorials.LinAlg import LinTran
            from OurManim import use_default_style
            tut = LinTran()
            tut.show()
            use_default_style()
        else:
            self._show_video('linear-transfomation.mp4')

    def _show_video(self, video_file):
        from Menus.GuiReferences import get_default_browser, get_preferable_browser, get_doc_folder_url
        browser = get_preferable_browser()
        video_url = get_doc_folder_url() + '/' + video_file
        browser.open_new(video_url)

    def show_matrix_inverse(self):
        self._show_video('matrix-inverse.mp4')

    def show_moore_penrose_inverse(self):
        self._show_video('moore-penrose-inverse.mp4')

    def show_affine_transformation(self):
        from molass_legacy.SerialAnalyzer.AffineDemo import AffineDemoDialog
        dialog = AffineDemoDialog( 'Affine transformation demo', parent=self.parent )
        dialog.show()

    def show_svd_animation(self):
        from Menus.GuiReferences import get_default_browser, get_preferable_browser, get_doc_folder_url
        browser = get_preferable_browser()
        movie_url = get_doc_folder_url() + '/SVD-animation.mp4'
        browser.open_new(movie_url)

    def show_svd_tutorial_2d(self):
        from Tutorials.SvdTutorial2D import SvdTutorial2D
        tut = SvdTutorial2D(self.parent)
        tut.show()

    def show_svd_tutorial_3d(self):
        from Tutorials.SvdTutorial3D import SvdTutorial3D
        tut = SvdTutorial3D(self.parent)
        tut.show()

    def show_mf_tutorial(self):
        from Tutorials.MatrixFactorization import MatrixFactorization
        tut = MatrixFactorization(self.parent)
        tut.show()

    def show_conjugate_gradient(self):
        import Extrapolation
        from ConjugateGradientDemo import ConjugateGradientDemo
        from molass_legacy.Mapping.MapperConstructor import create_mapper
        from molass_legacy._MOLASS.SerialSettings import dump_settings

        # dump_settings()
        parent = self.parent
        parent.config( cursor='wait' )
        parent.update()
        analysis_copy = parent.pre_recog.get_analysis_copy()

        mapper = create_mapper( parent, analysis_copy, analyzer_dialog=parent, logger=parent.tmp_logger )
        conc_factor = 5
        analysis_copy.set_mc_vector(mapper, conc_factor)        # TODO: set_mc_vector

        in_folder = parent.in_folder.get()
        try:
            demo = ConjugateGradientDemo(parent, in_folder, analysis_copy, mapper)
            parent.config( cursor='' )
            demo.show()
        except Exception as exc:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            MessageBox.showerror("ERROR", "Demo failed with the following error. Try another set of data.\n\n\n----" + str(etb), parent=parent)



# coding: utf-8
"""

    DataModelsAgg.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""

import numpy                as np


def _GuinierPorodAgg( C, Rg, d, x ):

    y = np.zeros( (len(x),) )
    for i in range(2):
        C_  = C[i]
        Rg_ = Rg[i]
        d_  = d[i]

        Q1  = 1/Rg_ * np.sqrt( 3*d_/2 )
        D   = C_ * np.exp( -d_/2 ) * np.power( 3*d_/2, d_/2 ) / np.power( Rg_, d_ )
        y_  = np.hstack( [  C_ * np.exp( - x[ x < Q1 ]**2 * Rg_**2 / 3 ),
                            D  / np.power( x[ x >= Q1 ], d_ )
                         ] )
        y += y_
    return y

class GuinierPorodAgg:
    def __init__( self, C, Rg, d ):
        self.C  = C
        self.Rg = Rg
        self.d  = d

    def __call__( self, x, sigma=0 ):
        # TODO: sigma=0 to be deleted
        return _GuinierPorodAgg( self.C, self.Rg, self.d, x )




# coding: utf-8
"""

    OurToplevel.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF

"""

import tkinter as Tk

class OurToplevel(Tk.Toplevel):
    def __init__(self, parent, title):
        Tk.Toplevel.__init__(self, parent)
        self.wm_title(title)
        body = Tk.Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)
        self.buttonbox()
        rootx = parent.winfo_rootx()
        rooty = parent.winfo_rooty()
        self.geometry("+%d+%d" % (rootx+50,rooty+50))
        self.update()
        self.protocol( "WM_DELETE_WINDOW", self.close )



# coding: utf-8
"""
    SaxsSimulator.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
import matplotlib.cm as cm
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from OurToplevel import OurToplevel
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color, MplBackGround, reset_to_default_style
from OurManim import manim_init, use_default_style, Animation, Collection, TextGroup, Parallelogram, rotation, angle
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from .EdPlotter import get_name_for_title, EdPlotter
from .SaxsDetector import SaxsDetector
from .SaxsCurve import SaxsCurve
from molass_legacy.KekLib.NumpyUtils import np_loadtxt_robust

class SaxsSimulator(OurToplevel):
    def __init__(self, parent, path=None, data=None, save_button=False, animate=False, on_close=None):
        self.parent = parent
        self.path = path
        self.save_button = save_button
        self.animate = animate
        self.data = data
        self.on_close = on_close
        self.esc = None
        self.popup_menu = None
        OurToplevel.__init__(self, parent, "SAXS Simulator")

    def close(self):
        self.destroy()
        if self.on_close is not None:
            self.on_close()

    def show(self):
        # dummy for compatibility with Dialog
        pass

    def body(self, body_frame):
        tk_set_icon_portable(self)
        self.mpl_bg = MplBackGround()
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=1)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        self.fig  = plt.figure(figsize=(21,7))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_mpl_button_press)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.build_control_panel(pframe)
        self.draw(self.path)

        self.add_dnd_bind()

    def add_dnd_bind(self):
        self.mpl_canvas_widget.register_drop_target("*")

        def dnd_handler(event):
            self.on_entry(event)

        self.mpl_canvas_widget.bind("<<Drop>>", dnd_handler)

    def build_control_panel(self, pframe):
        pass

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        if self.save_button:
            w = Tk.Button(box, text="Save", width=10, command=self.save)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

    def ok(self):
        self.close()

    def draw(self, file=None):
        fig = self.fig
        fig.clf()

        if file is None:
            from_name = ''
        else:
            from_name = ' from ' + get_name_for_title(file)
        fig.suptitle('SAXS Simulation' + from_name, fontsize=24)
        ax1 = fig.add_subplot(131, projection='3d')
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        self.axes = [ax1, ax2, ax3]
        fig.tight_layout()
        fig.subplots_adjust(top=0.85, wspace=0.15, bottom=0.1)

        if file is None and self.data is None:
            self.draw_blank()
        else:
            q, in_y = self.get_input_curve_data(file)

            if self.data is None:
                import mrcfile
                with mrcfile.open(file) as mrc:
                    data = mrc.data
            else:
                data = self.data

            cmap = cm.plasma
            denss_results = file is not None and os.path.exists(file)
            self.esc = EdPlotter(fig, ax1, data, cmap, file, denss_results=denss_results)

            if self.animate:
                # call self.esc.make_anim with after
                # to release the D&D source window
                self.after(0, lambda: self.esc.make_anim(random=True))

            self.detector = SaxsDetector(fig, ax3, data, q, in_y)
            r_y = self.detector.curve_y
            self.curve = SaxsCurve(fig, ax2, q, in_y, r_y)
        self.mpl_canvas.draw()

    def draw_blank(self):
        fig = self.fig
        for ax in self.axes:
            ax.set_axis_off()
        ax2 = self.axes[1]
        ax2.set_xlim(0,1)
        ax2.text(0.5, 0.5, "Drag and drop\nan mrc file to view.", ha='center', fontsize=50, alpha=0.5)
        fig.tight_layout()

    def on_entry(self, event):
        files = event.data.split(' ')
        print('on_entry:', files)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw(files[0])
        reset_to_default_style()

    def on_mpl_button_press(self, event):
        if self.esc is None:
            if event.button == 3:
                self.create_popup_menu()
                w, h, x, y = split_geometry(self.geometry())
                self.popup_menu.post(x + event.x + 20, y + h - event.y - 50)
            return

        self.esc.on_click(event)

    def create_popup_menu(self):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu( self, tearoff=0 )
            self.popup_menu.add_command( label='Ball', command=self.show_ball )
            self.popup_menu.add_command( label='Ellipsoid', command=self.show_ellipsoid )
            self.popup_menu.add_command( label='Disc', command=self.show_disc )
            self.popup_menu.add_command( label='Rod', command=self.show_rod )
            self.popup_menu.add_command( label='Torus', command=self.show_torus )
            self.popup_menu.add_command( label='from PDB', command=self.show_pdb_dialog)

    def get_input_curve_data(self, file):
        if file is None:
            file_exists = False
        else:
            path = file.replace('.mrc', '.dat')
            if os.path.exists(path):
                array, comments = np_loadtxt_robust(path)
                q = array[:,0]
                y = array[:,1]
                file_exists = True
            else:
                file_exists = False
        if not file_exists:
            # TODO: make an appropriate q
            q = np.linspace(0, 0.5, 600)
            y = None
        return q, y

    def save(self):
        pass

    def show_ball(self):
        from molass_legacy.Saxs.SaxsSamples import BallVoxels
        ball = BallVoxels(radius=0.2)
        self.path = "a ball"
        self.data = ball.get_data(density=0.1)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw()
        reset_to_default_style()

    def show_ellipsoid(self):
        from molass_legacy.Saxs.SaxsSamples import EllipsoidVoxels
        ball = EllipsoidVoxels(radii=(0.24, 0.2, 0.16))
        self.path = "an ellipsoid"
        self.data = ball.get_data(density=0.1)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw()
        reset_to_default_style()

    def show_disc(self):
        from molass_legacy.Saxs.SaxsSamples import DiscVoxels
        disc = DiscVoxels(radius=0.3)
        self.path = "a disc"
        self.data = disc.get_data(density=0.1)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw()
        reset_to_default_style()

    def show_rod(self):
        from molass_legacy.Saxs.SaxsSamples import DiscVoxels
        disc = DiscVoxels(radius=0.05, height=0.5)
        self.path = "a rod"
        self.data = disc.get_data(density=0.1)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw()
        reset_to_default_style()

    def show_torus(self):
        from molass_legacy.Saxs.SaxsSamples import TorusVoxels
        torus = TorusVoxels(R=0.6, r=0.2)
        self.path = "a torus"
        self.data = torus.get_data(density=0.1)
        self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
        self.draw()
        reset_to_default_style()

    def show_pdb_dialog(self):
        from molass_legacy._MOLASS.SerialSettings import get_setting

        analysis_folder = get_setting('analysis_folder')
        if analysis_folder is None:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            MessageBox.showinfo(
                'Retry Suggestion',
                'This operation requires an output folder.\n'
                'Please retry after specifying "Analysis Result Folder"\n'
                'in the Main Dialog.\n',
                parent=self,
                )
            return

        from Pdb.PdbFetcherDialog import PdbFetcherDialog
        dialog = PdbFetcherDialog(self)
        dialog.show()
        if dialog.applied:
            self.mpl_bg = MplBackGround()   # required when Toplevel instead of Dialog
            mrc_file = dialog.get_mrc_file_path()
            self.draw(mrc_file)
            reset_to_default_style()



"""
    mplclofa.py

    inspired by https://github.com/jwkvam/celluloid

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.animation as animation


class MplClone:
    def __init__(self, camera, obj):
        self.camera = camera
        self.obj = obj
        self.legend_obj = None
        self.legend_labels = []

    def set_title(self, title):
        # using text method instead of set_title, since the latter seems not ok for animation
        artist = self.obj.text(0.5, 1.05, title, transform=self.obj.transAxes, ha="center")
        self._append(artist)

    def plot(self, *args, **kwargs):
        label = kwargs.get("label", None)
        self.legend_labels.append(label)
        artist, = self.obj.plot(*args, **kwargs)
        self._append(artist)

    def legend(self):
        if self.legend_obj is None:
            self.legend_obj = self.obj.legend()

    def _append(self, artist):
        self.camera.frame_artists.append(artist)

    def _update_legend(self):
        if self.legend_obj is None:
            return

        texts = self.legend_obj.get_texts()
        for text, label in zip(texts, self.legend_labels):
            print("label=", label)
            text.set_text(label)
            self._append(text)

        self.legend_labels = []

class Camera:
    def __init__(self, fig):
        self.fig = fig
        self.photos = []
        self.frame_artists = []
        self.legends = []

    def clone(self, *args):
        ret_objects = []
        for obj in args:
            ret_objects.append(MplClone(self, obj))
        self.cloned = ret_objects
        return ret_objects

    def snap(self):
        for clone in self.cloned:
            clone._update_legend()
        self.photos.append(self.frame_artists)
        self.frame_artists = []
        self.legends = []

    def animate(self, **kwargs):
        anim = animation.ArtistAnimation(self.fig, self.photos, **kwargs)
        return anim

def demo():

    fig, axes_ = plt.subplots(nrows=2)
    fig.subplots_adjust(hspace=0.5)

    camera = Camera(fig)

    ax1, ax2 = camera.clone(*axes_)

    fig.suptitle("mplclofa demo")

    t = np.linspace(0, 2 * np.pi, 32, endpoint=False)
    for i in t:

        ax1.set_title("ax1: %.3g" % i)
        ax2.set_title("ax2: %.3g" % i)

        ax1.plot(t, np.sin(t + i), color='blue', label="np.sin(t + %.3g)" % i)
        ax2.plot(t, np.sin(t - i), color='blue', label="np.sin(t - %.3g)" % i)

        ax1.legend()    # not yet updating labels
        ax2.legend()    # not yet updating labels

        camera.snap()

    anim = camera.animate()
    plt.show()

if __name__ == '__main__':
    demo()



# coding: utf-8
"""
    SvdAnimation.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color
from OurManim import manim_init, use_default_style, Animation, Collection, TextGroup, Parallelogram, rotation, angle
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class SvdAnimation(Dialog):
    def __init__(self, parent=None):
        if parent is None:
            from SerialTestUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        self.num_frames = 100
        manim_init()
        Dialog.__init__(self, parent, "SVD Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig, self.axes = plt.subplots(nrows=1, ncols=2, figsize=(16,8))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        self.draw()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Save", width=10, command=self.save)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw(self):
        ax1, ax2 = self.axes

        square_labels = ['(0,0)', '(1,0)', '(1,1)', '(0,1)']
        square_points = [(0,0), (1,0), (1,1), (0,1)]
        data_labels = ['(a,b)', '(a+c, b+d)', '(c,d)']

        R1 = rotation(-np.pi/4)
        SG = np.diag([4, 1.5])
        R2 = rotation(np.pi/6)
        for x in [R1, SG, R2]:
            print(x)
        F = R1@SG@R2
        print('F=', F)
        f1 = F[:,0]
        f2 = F[:,1]
        data_points = [f1, f1+f2, f2]

        labels = square_labels + data_labels
        points = square_points + data_points
        d = 0.05

        for ax, t in zip(self.axes, [" (cumulative)", " (separative)"]):
            ax.set_title("SVD animation" + t, fontsize=30)
            ax.set_xlim(-4.0, 4.0)
            ax.set_ylim(-4.0, 4.0)
            ax.set_aspect('equal')
            if ax == self.axes[0]:
                labels_ = labels
                points_ = points
            else:
                labels_ = square_labels
                points_ = square_points
            for t, p in zip(labels_, points_):
                ax.plot(*p,'o')
                ax.text(p[0]+d, p[1]+d, t, fontsize=16)
            square =  Polygon(square_points, alpha=0.3, color='pink')
            ax.add_patch(square)

        collection = Collection(self.num_frames)

        final_target = [(0,0)] + data_points
        print('final_target=', final_target)

        para1 = Parallelogram(ax1, square_points, alpha=0.3, color='cyan')
        para1.set_target_vertices(final_target)
        anim_1 = Animation(0, 20)
        anim_1.append(para1)
        collection.append(anim_1)

        M = np.array([ final_target[1], final_target[3] ]).T 
        print('M=', M)
        svd = np.linalg.svd(M)
        for x in svd:
            print(x)

        # U, s, VT = svd
        U, s, VT = R1, np.diag(SG), R2

        v1 = VT[:,0]
        v2 = VT[:,1]
        VT_points = [(0,0), v1, v1+v2, v2]
        para2 = Parallelogram(ax1, square_points, motion='rotation', alpha=0.3, color='cyan')
        para2.set_target_vertices(VT_points)
        para2_ = Parallelogram(ax2, square_points, motion='rotation', alpha=0.3, color='cyan')
        para2_.set_target_vertices(VT_points)
        anim_2 = Animation(20, 40)
        anim_2.append(para2)
        anim_2.append(para2_)
        collection.append(anim_2)

        sVT = np.dot(np.diag(s), VT)
        sv1 = sVT[:,0]
        sv2 = sVT[:,1]
        sVT_points = [(0,0), sv1, sv1+sv2, sv2]
        para3 = Parallelogram(ax1, VT_points, alpha=0.3, color='cyan')
        para3.set_target_vertices(sVT_points)
        sv1_ = SG[:,0]
        sv2_ = SG[:,1]
        SG_points = [(0,0), sv1_, sv1_+sv2_, sv2_]
        para3_ = Parallelogram(ax2, square_points, alpha=0.3, color='cyan')
        para3_.set_target_vertices(SG_points)
        anim_3 = Animation(40, 60)
        anim_3.append(para3)
        anim_3.append(para3_)
        collection.append(anim_3)

        para4 = Parallelogram(ax1, sVT_points, motion='rotation', alpha=0.3, color='cyan')
        para4.set_target_vertices(final_target)
        u1 = U[:,0]
        u2 = U[:,1]
        U_points = [(0,0), u1, u1+u2, u2]
        para4_ = Parallelogram(ax2, square_points, motion='rotation', alpha=0.3, color='cyan')
        para4_.set_target_vertices(U_points)
        anim_4 = Animation(60, 80)
        anim_4.append(para4)
        anim_4.append(para4_)
        collection.append(anim_4)

        self.anim = collection.make_animation(self.fig)

    def save(self):
        import os
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        path = os.path.join(os.getcwd(), "anim.mp4")
        self.anim.save(path, writer="ffmpeg")
        MessageBox.showinfo("Save Notification", 'The movie has been saved to "%s"' % path, parent=self)



"""
Beautiful equation - Euler's identity

The code prints the mathematics 'beautiful equation' or 'Euler's identity' result.
The formula is:

        e**πi + 1 = 0

where:
e - Euler's number, 2.718...
π - pi, 3.14159...
i - imaginary number, equal to √-1 (i**2 = -1)

The formula is called 'beautiful' because it uses the mathematics most widely important constants:
e - the base of natural logarithms
π - the ratio of circumference to diameter
i - imaginary unit of complex numbers
0 - additive identity (n + 0 = n)
1 - multiplicative identity (n * 1 = n)

"""

import math
print (f"{(math.e**(math.pi*1j)) + 1:.12f}")



import numpy                as np
import matplotlib.pyplot    as plt
import statsmodels.api      as sm
from scipy.stats            import norm

def regress( N, A, B, E ):
    x = np.linspace( 0, 1, N )
    y0  = A + B*x
    y_  = y0 + E * np.random.randn( N )
    X = sm.add_constant( x )
    result = sm.OLS( y_, X ).fit()
    return result

A   = 5
B   = 3
E   = 0.2

N_regress = 100

fig = plt.figure( figsize=( 12, 9 ) )

varnames = [ 'A', 'B' ]
midvals = [ A, B ]

xlims = []

for n, num_regress in enumerate( [ 10, 100, 1000 ] ):
    ab_error = np.sqrt( np.diag( regress( num_regress, A, B, E ).cov_params() ) )
    print( 'sqrt( cov_params )=', ab_error )
    result_array = [ regress( num_regress, A, B, E ) for i in range(1000) ]
    ab_array = np.array( [ result.params for result in result_array ] )
    ab_error_ = np.average( np.array( [ np.sqrt( np.diag( result.cov_params() ) ) for result in result_array ] ), axis=0 )
    print( 'averaged sqrt( cov_params )=', ab_error_ )

    print( 'simulated stdev=', np.std( ab_array, axis=0 ) )

    for i in range(2):
        ax = fig.add_subplot( 3, 2, n*2+i+1 )
        ax.set_title( 'distribution of estimated %s and normal curve when N=%d' % ( varnames[i], num_regress ) )
        ax.hist( ab_array[:, i], bins=20, normed=True )
        if n == 0:
            xmin, xmax = ax.get_xlim()
            xlims.append( [xmin, xmax] )
        else:
            xmin, xmax = xlims[i]
        ax.set_xlim( xmin, xmax )
        x = np.linspace( xmin, xmax, 100 )
        error = ab_error_[i]
        y = norm.pdf( x, loc=midvals[i], scale=error )
        ax.plot( x, y )
        ymin, ymax = ax.get_ylim()
        ax.text( xmin+0.02*(xmax-xmin), ymax*0.8,
                'error level=%.3g\nerror level / sqrt(%d)=%.3g\nerror=%.3g'
                % ( E, num_regress, E/np.sqrt(num_regress), error ) )

plt.tight_layout()
plt.show()



"""

    GuinierAnalysisResultBook.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
import numpy as np
import logging
from openpyxl.chart import BarChart, LineChart, ScatterChart, Reference, Series
from openpyxl.chart.series_factory  import SeriesFactory
from openpyxl.chart.error_bar import ErrorBars
from openpyxl.chart.data_source import NumDataSource, NumData, NumVal
from openpyxl.chart.layout import Layout, ManualLayout
from molass_legacy.KekLib.OpenPyXlUtil import save_allowing_user_reply, LINE_WIDTH
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy.KekLib.HtmlColorNames import *
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.Reports.DefaultFont import set_default_font
set_default_font()

CHART_WIDTH     = 54
CHART_ADJUST    = 0.005

COLNAMES = [
    '№',               #  0 A
    'folder',           #  1 B
    'file',             #  2 C
    'concentration',    #  3 D
    'basic_quality',    #  4 E
    'positive_score',   #  5 F
    'stdev_score',      #  6 G      row[5] end_consistency
    'q_rg_score',       #  7 H      row[6] stdev_score
    'end_consistency',  #  8 I      row[7] q_rg_score
    'atsas_qualiity',   #  9 J      row[8] atsas_qualiity
    'molass.I(0)',      # 10 K
    'gp.I(0)',          # 11 L
    'atsas.I(0)',       # 12 M
    'molass.Rg',        # 13 N
    'gp.Rg',            # 14 O
    'atsas.Rg',         # 15 P
    'molass.I(0)/conc.',# 16 Q
    'gp.I(0)/conc.',    # 17 R
    'atsas.I(0)/conc.', # 18 S
    ]


class GuinierAnalysisResultBook:
    def __init__( self, wb, ws, array, j0, parent=None ):

        self.logger  = logging.getLogger( __name__  )
        self.parent  = parent   # SerialController
        self.wb_path = None
        self.wb = wb
        self.ws = ws
        self.j0 = j0
        ws.title = 'Guinier Analysis'
        self.num_points = len( array )
        self.excel_client_ = None
        self.include_end_consistency = True

        colnames_ = list( COLNAMES )
        if self.include_end_consistency:
            for k in [ 17, 14, 11 ]:
                del colnames_[k]
        else:
            colnames_[8] = COLNAMES[9]
            colnames_[9] = COLNAMES[8]

        ws.append( colnames_ )

        Iz_array = []
        Izc_array = []
        Rg_array = []
        i = 1
        for row in array:
            i += 1
            row_ = list(row)
            if self.include_end_consistency:
                row_[7] = row[5]        # move end_consistency
                row_[5:7] = row[6:8]    # shift stdev_score, q_rg_score
                for k in [ 13, 10 ]:
                    del row_[k]
                fmt_cols = [ 'K', 'L' ]
            else:
                row_[8] = row[5]        # move end_consistency
                row_[5:8] = row[6:9]    # shift stdev_score, q_rg_score, atsas_qualiity
                fmt_cols = [ 'K', 'L', 'M' ]

            fmt_rows =  [ '=IF(D{0}=0,0,{1}{0}/D{0})'.format( i, c ) for c in fmt_cols ]

            ws.append( [j0+i-2] + row_ + fmt_rows )
            # autorg_kek is excluded for it may give extraordinary values for some data
            Iz_array.append( max( row[9:11] ) )     # 9:11 => [ kek.I0, fit.I0 ]
            Izc_array.append( max( row[9:11] )/row[2]  )
            Rg_array.append( row[12] )

        self.iz_array   = np.array(Iz_array)
        self.Izc_array  = np.array(Izc_array)
        self.rg_array   = np.array(Rg_array)

        self.xvalues    = Reference(ws, min_col=1, min_row=2, max_row=len(array)+1)

        if self.include_end_consistency:
            chart_start_col = 'R'
        else:
            chart_start_col = 'U'

        if get_setting( 'use_xray_conc' ) == 0:
            xray_text = ''
        else:
            xray_text = ' (Xray-proportional'
            if get_dev_setting( 'smoothed_xray_conc' ) == 1:
                xray_text += ' & smoothed'
            xray_text += ')'
        self.create_ConcentrationChart( array, ws,       chart_start_col+'2', 'Concentration' + xray_text, 4 )
        self.create_QualityChart( array, ws,            chart_start_col+'19' )
        self.create_IzChart( array, ws,                  chart_start_col+'36' )
        self.create_Iz_devided_by_C_Chart( array, ws,    chart_start_col+'53' )
        self.create_RgChart( array, ws,                  chart_start_col+'70' )

    def create_ConcentrationChart( self, array, ws, pos, title, col ):

        c_ = LineChart()
        c_.title = title
        c_.style = 13
        c_.y_axis.title = 'Concentration'
        c_.x_axis.title = 'Elution №'

        num_rows = len(array)
        data = Reference(ws, min_col=col, max_col=col, min_row=1, max_row=num_rows+1 )
        c_.add_data(data, titles_from_data=True)
        c_.set_categories(self.xvalues)
        c_.width    = CHART_WIDTH
        c_.hight    = 15

        colors = [ steelblue ]

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            s_.graphicalProperties.line.width               = LINE_WIDTH    # in EMUs

        self.set_layout( c_, adjust=0 )
        ws.add_chart(c_, pos)

    def create_QualityChart( self, array, ws, pos ):

        c_ = BarChart()
        c_.title = "Quality Factors"
        c_.type  = "col"
        c_.style = 10
        c_.grouping = "stacked"
        c_.overlap = 100
        # c_.gapWidth = 300
        c_.y_axis.title = 'Quality'
        c_.x_axis.title = 'Elution №'

        num_rows = len(array)
        max_col = 10 if self.include_end_consistency else 9
        data = Reference(ws, min_col=5, max_col=max_col, min_row=1, max_row=num_rows+1 )
        c_.add_data(data, titles_from_data=True)
        c_.set_categories(self.xvalues)
        c_.width    = CHART_WIDTH
        c_.hight    = 15

        # colors = [ lightskyblue, darkorange, gold, dodgerblue, limegreen, red ]
        # remove end_consistency
        colors = [ lightskyblue, darkorange, dodgerblue, limegreen, red ]
        if self.include_end_consistency:
            colors.insert( 4, gold )

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            s_.graphicalProperties.solidFill                = color
            s_.graphicalProperties.line.solidFill           = color
            # the last series, i.e. atsas quality, which is the fitfh FullSeriesCollection,
            # will be modified using COM in self.add_annonations_impl method.

        self.set_layout( c_, adjust=0 )
        ws.add_chart(c_, pos)

    def create_IzChart( self, array, ws, pos ):
        c_ = LineChart()
        c_.title = 'I(0)'
        c_.style = 13
        c_.y_axis.title = 'I(0)'
        c_.x_axis.title = 'Elution №'

        if self.include_end_consistency:
            min_col=11
            max_col=12
            colors = [ steelblue, darkorange ]
        else:
            min_col=11
            max_col=13
            colors = [ steelblue, darkgray, darkorange ]

        num_rows = len(array)
        data = Reference(ws, min_col=min_col, max_col=max_col, min_row=1, max_row=num_rows+1 )
        c_.add_data(data, titles_from_data=True)
        c_.set_categories(self.xvalues)
        c_.width    = CHART_WIDTH
        c_.hight    = 15

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            s_.graphicalProperties.line.width               = LINE_WIDTH    # in EMUs

        self.set_layout( c_, adjust=CHART_ADJUST )
        ws.add_chart(c_, pos)

    def create_Iz_devided_by_C_Chart( self, array, ws, pos ):

        c_ = LineChart()
        c_.title = "I(0) / Concentration"
        c_.style = 13
        c_.y_axis.title = 'I(0) / Conc.'
        c_.x_axis.title = 'Elution №'

        if self.include_end_consistency:
            min_col=15
            max_col=16
            colors = [ steelblue, darkorange ]
        else:
            min_col=17
            max_col=19
            colors = [ steelblue, darkgray, darkorange ]

        num_rows = len(array)
        data = Reference(ws, min_col=min_col, max_col=max_col, min_row=1, max_row=num_rows+1 )
        c_.add_data(data, titles_from_data=True)
        c_.set_categories(self.xvalues)
        c_.width    = CHART_WIDTH
        c_.hight    = 15

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            s_.graphicalProperties.line.width               = LINE_WIDTH    # in EMUs

        self.set_layout( c_, adjust=CHART_ADJUST*4.6 )
        ws.add_chart(c_, pos)

    def create_RgChart( self, array, ws, pos ):

        c_ = LineChart()
        c_.title = 'Rg'
        c_.style = 13
        c_.y_axis.title = 'Rg'
        c_.x_axis.title = 'Elution №'

        if self.include_end_consistency:
            min_col=13
            max_col=14
            colors = [ steelblue, darkorange ]
        else:
            min_col=14
            max_col=16
            colors = [ steelblue, darkgray, darkorange ]

        num_rows = len(array)
        data = Reference(ws, min_col=min_col, max_col=max_col, min_row=1, max_row=num_rows+1 )
        c_.add_data(data, titles_from_data=True)
        c_.set_categories(self.xvalues)
        c_.width    = CHART_WIDTH
        c_.hight    = 15

        # Style the lines

        for i, color in enumerate(colors):
            s_ = c_.series[i]
            s_.marker.symbol = "circle"
            s_.marker.graphicalProperties.solidFill         = color
            s_.marker.graphicalProperties.line.solidFill    = color
            s_.graphicalProperties.line.solidFill           = color
            s_.graphicalProperties.line.width               = LINE_WIDTH    # in EMUs

        self.set_layout( c_, adjust=CHART_ADJUST*0.4 )
        ws.add_chart(c_, pos)

    def set_layout( self, chart, adjust=0 ):
        chart.layout = Layout(
            ManualLayout(
            x=0.04-adjust, y=0.1,
            h=0.8,  w=0.895+adjust,
            xMode="edge",
            yMode="edge",
            )
        )

    def add_annonations(self, book_file, ranges, debug=False):
        from .GuinierExcelFormatter import GuinierReportArgs
        args = GuinierReportArgs(self.ws.title, self, book_file, ranges)    # make it picklable

        if self.parent.more_multicore:
            self.add_annonations_more_multicore(args)
        else:
            self.add_annonations_less_multicore(args, debug)

    def add_annonations_more_multicore(self, args):
        self.parent.teller.tell('guinier_book', args=args)

    def add_annonations_less_multicore(self, args, debug):
        from .GuinierExcelFormatter import add_guinier_annonations

        try:
            add_guinier_annonations(self.parent.excel_client, args, self.logger, debug=debug)
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.logger.warning( etb )

    def save( self, xlsx_file ):
        save_allowing_user_reply( self.wb, xlsx_file )



# -*- coding: utf-8 -*-
"""

    ファイル名：   PilatusImageInfoTable.py

    処理内容：

       Pilatus 画像データ情報の一覧

"""
from __future__ import division, print_function, unicode_literals

import sys
import copy
import logging
import traceback
from datetime               import datetime

from molass_legacy.KekLib.OurTkinter             import Tk, ToolTip
import OurMessageBox        as     MessageBox 
from CuntomizedPandasTable import PandasTable

from PilatusUtils           import get_data_info
from GuiPreferences         import PreferencesDialog
from PilatusImageProperty   import ImagePropertyWindow
from Preferences            import get_preference, temporary_preferences_begin, temporary_preferences_end, get_usual_preference
from SynthesizerSettings    import get_setting

# mapping displayed table col to internal array col
tif_dir_map = {
        0   :   'in_folder',
        1   :   'in_folder',
        4   :   'in_folder',
        7   :   'syn_folder',
        }

class PilatusImageInfoTable:
    def __init__( self, parent, synthesizer, suggestion ):
        self.parent     = parent
        self.var        = None
        self.table      = None
        self.scrollbar  = None
        self.current_row = 0
        self.synthesizer = synthesizer
        self.x_view     = None
        self.y_view     = None
        self.suggestion = suggestion
        self.temporary_i_ratio_array = None
        self.num_selected_rows  = 0
        self.current_data_end = 0

        self.refresh()

    def refresh( self, restore_view=False, pilatus_counter=None, bottom_view=False, log_file_path=None ):

        # print( 'refresh: bottom_view=', bottom_view )

        if not restore_view:
            self.temporary_i_ratio_array = None

        self.data_array = []

        in_folder   = get_setting( 'in_folder' )
        adj_folder  = get_setting( 'adj_folder' )
        syn_folder  = get_setting( 'syn_folder' )

        counter_id = get_preference( 'detection_counter' )
        # print( 'counter_id=', counter_id )

        try:
            _, _, data_array, pilatus_counter = get_data_info( in_folder, adj_folder, syn_folder, pilatus_counter, counter_id,
                                                        log_file_path=log_file_path
                                                        )
        except:
            logging.exception( 'Unexpected error' )
            MessageBox.showerror( 'Unexpected Error', traceback.format_exc() )
            return

        self.data_array = data_array
        self.pilatus_counter = pilatus_counter

        if self.current_data_end > len( data_array ):
            # 作動中に出力フォルダのデータを削除した場合などが該当する。
            self.current_data_end = 0

        columns = [
                [ 'Original Image', 140 ],
                [ 'Shited Image 1', 140 ],
                [ 'Relative Position 1', 100 ],
                [ 'Intensity Ratio 1 None', 110 ],
                [ 'Shited Image 2', 140 ],
                [ 'Relative Position 2', 100 ],
                [ 'Intensity Ratio 2 None', 110 ],
                [ 'Synthesized Image', 140 ],
            ]

        num_cols = 8

        row = 0
        self.var = []

        for row_rec in self.data_array:
            row += 1
            # print( '[%d]' % (row), row_rec )

            var_rec = []

            sample_id   = row_rec[0]
            i = 0
            for sub_rec in row_rec[1]:
                i += 1

                var_rec.append( sub_rec[0] )

                if i == 1:
                    continue

                disp_value = '%s,%s' % tuple( sub_rec[1] )
                var_rec.append( disp_value )

                i_ratio_value = sub_rec[2]
                if i_ratio_value:
                    changed_counter_id = ''
                    if self.temporary_i_ratio_array:
                        # TODO
                        pass
                    disp_value = '%.5f%s' % ( i_ratio_value, changed_counter_id )   # 表示用の桁数調整
                else:
                    disp_value = ''
                var_rec.append( disp_value )

                if i == len( row_rec[1] ):
                    for j in range( len( var_rec ), num_cols-1 ):
                        var_rec.append( '' )
                    var_rec.append( sub_rec[4] )

            self.var.append( var_rec )

        if self.table:
            # self.save_xy_views()
            # self.table.destroy()
            if self.scrollbar:
                self.scrollbar.destroy()

        created = False
        if self.table == None:
            self.table = PandasTable( self.parent, columns=columns )
            self.table.pack( expand=1, fill=Tk.BOTH )
            created = True

        self.table.import_array( self.var )

        if created:
            # create a popup menu
            self.menu = Tk.Menu( self.table, tearoff=0 )
            self.menu.add_command( label='Show Original Images',    command=lambda: self.do_action( 1 ) )
            self.menu.add_command( label='Show Adjusted Images',    command=lambda: self.do_action( 2 ) )
            self.menu.add_command( label='Make Synthsized Images',  command=lambda: self.do_action( 3 ) )
            self.menu_c = Tk.Menu( self.table, tearoff=0 )
            # ToolTip( self.menu_c, 'With the cascading submenus, you can temporarily change the preferences and execute.')
            # does not work expectedly
            self.menu.add_cascade( label='Execute with Temporary Preference Changes', menu=self.menu_c )
            self.menu_c.add_command( label='Show Original Simply',   command=lambda: self.do_action( 1, True ) )
            self.menu_c.add_command( label='Show Adjusted Images',   command=lambda: self.do_action( 2, True ) )
            self.menu_c.add_command( label='Make Synthsized Images', command=lambda: self.do_action( 3, True ) )
            self.menu.add_command( label='Tiff File Description',    command=self.show_properties )

            self.table.bindActionMenu( "<Button-3>", self.popup, corner=True )
            self.table.bindActionMenu( "<Double-Button-1>", lambda evet: self.do_action( 1 ), corner=False )

            ToolTip( self.table, '<Double-click> a row to show images. Select and <right-click> to show the action menu. Click top-left corner "No" to select all rows.')

        self.num_selectable_cols = num_cols

        if restore_view:
            # self.restore_xy_views()
            pass

        if bottom_view:
            self.set_view_to_bottom()
            self.on_resize()

        self.select_added_rows()

    def on_resize( self, event=None ):
        pass

    def set_view_to_bottom( self ):
        # TODO:
        pass

    def popup( self, event ):
        self.popup_position = ( event.x, event.y )
        # print 'popup_position=', self.popup_position

        # 改造した Tktable.dll ならば継ぎの work-around は不要
        # self.table.configure( state='normal' )

        self.menu.post( event.x_root, event.y_root )

    def popup_state_reset( self, event ):
        # 改造した Tktable.dll ならば継ぎの work-around は不要
        # self.table.configure( state='disabled' )
        pass

    def test_cmd( self, event ):
        if event.i == 0:
            return '%i, %i' % (event.r, event.c)
        else:
            return 'set'

    def browsecmd( self, event ):
        pass

    def select_all_rows( self ):
        self.table.selectAll()
        self.selected_rows = self.table.selectedRows()
        self.num_selected_rows = len( self.selected_rows )

    def select_added_rows( self ):
        if len( self.data_array ) == 0: return

        if self.current_data_end < len( self.data_array ):
            print( 'select_added_rows: selectRows %d, %d' % ( self.current_data_end, len( self.data_array ) ) )
            self.table.selectRows( self.current_data_end, len( self.data_array ) )
            self.selected_rows = self.table.selectedRows()
        else:
            self.table.selectionClear()
            self.selected_rows = []
            pass

        self.num_selected_rows = len( self.selected_rows )

    def get_selected_indices( self ):
        # 手動の選択もあるので、取り直す。
        self.selected_rows = self.table.selectedRows()
        self.num_selected_rows = len( self.selected_rows )
        return self.selected_rows

    def select_data_array( self, selected_indices ):
        array = []
        for i in selected_indices:
            array.append( self.data_array[ i ] )
        return array

    def clear_selection( self, indices ):
        # TODO: clear only indices
        self.table.selectionClear()

    def do_action( self, action, change=False ):
        self.popup_state_reset( 0 )

        if self.suggestion and self.suggestion.is_blinking():
            MessageBox.showwarning( "Not Allowed", "You can't do any action until you refresh the image data infomation table." )
            return

        selected_indices = self.get_selected_indices()
        exec_array = self.select_data_array( selected_indices )

        counter_id_is_changed = False
        if change:
            temporary_preferences_begin()
            dialog = PreferencesDialog( self.parent, "Temporary Preferences", self.pilatus_counter,
                            action=action )
            if not dialog.applied:
                return

            counter_id = get_preference( 'detection_counter' )
            if counter_id != get_usual_preference( 'detection_counter' ):
                counter_id_is_changed = True
                counter_dict = self.pilatus_counter.get_counter_dict( counter_id )
                if not self.temporary_i_ratio_array:
                    self.temporary_i_ratio_array = []
                    for rec in self.data_array:
                        self.temporary_i_ratio_array.append( None )

        start = datetime.now()
        # print '%s start.' % ( start.strftime( '%Y-%m-%d %H:%M:%S' ) )

        self.synthesizer.execute( action, exec_array )

        end = datetime.now()
        t = end - start
        # print '%s end. took %d seconds' % ( end.strftime( '%Y-%m-%d %H:%M:%S' ), t.seconds )

        if change:
            temporary_preferences_end()

        if action == 3:
            self.update_current( selected_indices )
            self.refresh( restore_view=True )
            # TODO: update in synthesizer.single_exec

        return

    def update_current( self, selected_indices ):
        if len( selected_indices ) > 0:
            end_index = selected_indices[-1] + 1
            if end_index > self.current_data_end:
                self.current_data_end = end_index
                print( 'new self.current_data_end=', self.current_data_end )

    def dummy( self ):
        pass

    def show_properties( self ):
        self.popup_state_reset( 0 )
        row, col = self.table.selectedCell()
        if row == None:
            # this case do happen.
            MessageBox.showinfo( 'No Row Selected Notification', 'No row is selected. Select a row and retry.' )
            return

        if col in tif_dir_map:
            dirkey = tif_dir_map[col]
            filename = self.var[row][col]
            dirpath = get_setting( dirkey )
            path = dirpath + '/' + filename
            title = filename + ' Description'
            ImagePropertyWindow( self.parent, title, path )
        else:
            MessageBox.showinfo( 'Select Tiff File Notice', 'Select tiff file cell and retry.' )
        return

    def save_xy_views( self ):
        self.x_view = self.table.xview()
        self.y_view = self.table.yview()

    def restore_xy_views( self ):
        if self.x_view and self.y_view:
            self.table.xview_moveto( self.x_view[0] )
            self.table.yview_moveto( self.y_view[0] )



"""
    Batch.BatchMappingUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings, set_setting
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Batch.LiteBatch import LiteBatch     

class MappingBatch(LiteBatch):
    def __init__(self, mapping_src):
        super().__init__()
        self.mapping_src = mapping_src

    def modelled_peaks_method(self, uv_x, uv_y_, xr_x, xr_y_, debug=True):
        def show_mapped_result(debug=True):
            from importlib import reload
            import Batch.BatchMappingImpl
            reload(Batch.BatchMappingImpl)
            from molass_legacy.Batch.BatchMappingImpl import show_mapped_result_impl
            show_mapped_result_impl(self, uv_x, uv_y_, xr_x, xr_y_, debug=debug)

        if debug:
            extra_button_specs = [("Show Mapped Result", show_mapped_result)]
            with plt.Dp(extra_button_specs=extra_button_specs):
                fig, ax = plt.subplots()
                fig.suptitle("modelled_peaks_method")
                plt.show()
        else:
            show_mapped_result(debug=False)

def load_lrfsrc_with_mapping(in_folder, mapping_src, clear=True):
    if clear:
        clear_temporary_settings()
    set_setting('in_folder', in_folder)
    mbatch = MappingBatch(mapping_src)
    lrf_src = mbatch.get_lrf_source(in_folder=in_folder, use_mapping=True)
    return lrf_src


# coding: utf-8
"""
    ThreeStateDimerSolver.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import os
import time
import numpy as np
from scipy import optimize, stats
from lmfit import Parameters, minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.SciPyCookbook import smooth
from SimpleUnfolding import RT
from SvdDenoise import get_denoised_data

def compute_constant_k(G, m, x):
    return np.exp((-G + m*x)/RT)

def compute_C(params, x):
    G1, m1, G2, m2, Pt = params
    K1 = compute_constant_k(G1, m1, x)
    K2 = compute_constant_k(G2, m2, x)
    Fd = (-K1*K2 + np.sqrt((K1*K2)**2 + 8*(1+K1)*K1*K2*Pt))/(4*Pt*(1+K1))

    C = np.array([2*Pt*Fd**2/(K1*K2), 2*Pt*Fd**2/K2, Fd])
    return C

class ThreeStateCurves:
    def __init__(self, params, ab_params):
        self.params = params
        self.ab_params = ab_params

    def compute_yn(self, x):
        self.C = compute_C(self.params, x)
        an = self.ab_params[0]
        return an*self.C[0,:]

    def compute_yi(self, x):
        ai = self.ab_params[1]
        return ai*self.C[1,:]

    def compute_yu(self, x):
        au = self.ab_params[2]
        return au*self.C[2,:]

class ThreeStateDimerSolver:
    def __init__(self):
        pass

    def solve(self, data):
        print(data.shape)
        M = get_denoised_data(data, rank=3)
        x = np.arange(M.shape[1])

        def compute_norm(params):
            C = compute_C(params, x)
            Cpinv = np.linalg.pinv(C)
            P = M@Cpinv
            z = np.linalg.norm(P@C - M)
            return z

        def get_init_params(G):
            G_init = G
            mx1 = (x[0]*2 + x[-1])/3
            m1_init = G_init/mx1
            mx2 = (x[0] + x[-1]*2)/3
            m2_init = G_init/mx2
            return G_init, m1_init, m2_init

        t0 = time.time()
        min_norm = None
        opt_x = None
        for g in range(0, 6):
            G_init, m1_init, m2_init = get_init_params(g)
            try:
                res= optimize.minimize(compute_norm, [G_init, m1_init, G_init, m2_init, 1], bounds=((0,10),(0,1),(0,10),(0,1),(0,10)))
            except:
                print([g], 'error')
                continue
            norm = compute_norm(res.x)
            print([g], norm, res.x)
            if min_norm is None or norm < min_norm:
                min_norm = norm
                opt_x = res.x
        self.params = opt_x
        print("It took", time.time() - t0)

        self.M = M
        self.compute_components(x)

    def compute_components(self, x):
        C = compute_C(self.params, x)
        Cpinv = np.linalg.pinv(C)
        P = self.M@Cpinv
        self.C = C
        self.P = P

    def plot_components(self, ax, data, index, title=None, axt=None, start=0):
        x = np.arange(data.shape[1]) + start
        y = data[index,:]
        self.x = x
        self.y = y
        C = self.C
        pn = C[0,:]
        pi = C[1,:]
        pu = C[2,:]
        an, ai, au = self.P[index,:]

        yn = an*pn
        yi = ai*pi
        yu = au*pu

        if axt is None:
            axt = ax.twinx()
        axt.grid(False)

        if title is not None:
            ax.set_title(title, fontsize=16)
        ax.set_xlabel('Eno')
        ax.set_ylabel('Proportion')
        axt.set_ylabel('Intensity')

        ax.plot(x, pn, label='$P_n$')
        ax.plot(x, pi, label='$P_i$')
        ax.plot(x, pu, label='$P_u$')

        axt.plot(x, y, color='orange', label='given')
        axt.plot(x, yn, ':', label='$y_n$')
        axt.plot(x, yi, ':', label='$y_i$')
        axt.plot(x, yu, ':', label='$y_u$')
        axt.plot(x, yn+yi+yu, ':', color='orange', label='$y_n+y_i+y_u$')

        ax.legend(bbox_to_anchor=(0, 0.5), loc='center left', fontsize=16)
        axt.legend(loc='upper right', fontsize=16)

    def draw_components_with_bands(self, ax, data, index, title=None, start=0, detail=False):
        """
        detail=False is not used here. It is just for interface compatibility with TwoStateSolver
        """
        from matplotlib.patches import Polygon
        from molass_legacy.KekLib.OurMatplotlib import get_color
        from GeometryUtils import polygon_area_centroid

        x = np.arange(data.shape[1]) + start
        y = data[index,:]
        self.x = x
        self.y = y
        C = self.C
        pn = C[0,:]
        pi = C[1,:]
        pu = C[2,:]
        an, ai, au = self.P[index,:]

        ax.set_xlabel('Eno')
        ax.set_ylabel('Proportion')

        ax.plot(x, y, color='orange', label='given')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(0, ymax)

        yn = an*pn
        yi = ai*pi
        yu = au*pu

        poly_points = list( zip( x, yn ) ) + [ (x[-1], 0), (x[0], 0) ]
        n_color = get_color(0)
        n_band = Polygon( poly_points, color=n_color, alpha=0.2 )
        ax.add_patch(n_band)
        cp = polygon_area_centroid(n_band)
        ax.text(*cp, "N", color=n_color, ha='center', va='center', fontsize=50)

        yn_yi = yn+yi
        poly_points = list(zip( x, yn )) + list( reversed( list( zip( x, yn_yi ) ) ) )
        i_color = get_color(1)
        i_band = Polygon( poly_points, color=i_color, alpha=0.2 )
        ax.add_patch(i_band)
        cp = polygon_area_centroid(i_band)
        ax.text(*cp, "I", color=i_color, ha='center', va='center', fontsize=50)

        yn_yi_yu = yn_yi + yu
        poly_points = list(zip( x, yn_yi )) + list( reversed( list( zip( x, yn_yi_yu ) ) ) )
        u_color = get_color(2)
        u_band = Polygon( poly_points, color=u_color, alpha=0.2 )
        ax.add_patch(u_band)
        cp = polygon_area_centroid(u_band)
        ax.text(*cp, "U", color=u_color, ha='center', va='center', fontsize=50)

    def get_modeled_curves(self, index):
        curves = ThreeStateCurves(self.params, self.P[index,:])
        return curves.compute_yn, curves.compute_yi, curves.compute_yu

    def get_imermediate_topx(self):
        G2, m2 = self.params[2:4]
        return G2/m2



"""
    Optimizer.BH.Customize.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy._lib._util import check_random_state
from .BasinHopping import AdaptiveStepsize, RandomDisplacement

class DefaultTakeStep(AdaptiveStepsize):
    def __init__(self, stepsize=0.5, interval=50, disp=False, seed=None):

        # set up the np.random generator
        self.rng = check_random_state(seed)
        displace = RandomDisplacement(stepsize=stepsize, random_gen=self.rng)

        AdaptiveStepsize.__init__(self, displace, interval=interval, verbose=disp)

    def get_rng(self):
        return self.rng

class BoundedRandomDisplacement(RandomDisplacement):
    def __init__(self, **kwargs):
        RandomDisplacement.__init__(self, **kwargs)

class CustomTakeStep(AdaptiveStepsize):
    def __init__(self, stepsize=0.5, interval=50, disp=False, seed=None):

        # set up the np.random generator
        self.rng = check_random_state(seed)
        displace = RandomDisplacement(stepsize=stepsize, random_gen=self.rng)

        AdaptiveStepsize.__init__(self, displace, interval=interval, verbose=disp)

    def get_rng(self):
        return self.rng



# coding: utf-8
"""
    AbsorbanceSaveDialog.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
import os
import numpy                as np
from bisect                 import bisect_right
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy._MOLASS.SerialSettings         import get_setting
from molass_legacy.KekLib.TkUtils                import split_geometry
from molass_legacy.KekLib.TkCustomWidgets        import FileEntry
from molass_legacy.KekLib.NumpyUtils             import np_savetxt

class AbsorbanceSaveDialog( Dialog ):
    def __init__( self, parent ):
        self.parent = parent
        self.caller_module = get_caller_module( level=2 )

    def show( self ):
        title = "Absorbance Data Save"
        Dialog.__init__( self, self.parent, title, auto_geometry=False, geometry_cb=self.adjust_geometry )

    def adjust_geometry( self ):
        w, h, x, y = split_geometry( self.parent.geometry() )
        self.geometry("+%d+%d" % (self.parent.winfo_rootx() + w//2,
                                  self.parent.winfo_rooty() + int(h*0.7) ))

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        label_frame = Tk.Frame(body_frame)
        label_frame.pack( padx=50, pady=10 )

        guide = Tk.Label(label_frame, text='Select what and where (file name) to save and press "OK"')
        guide.pack()

        detail_frame = Tk.Frame(body_frame)
        detail_frame.pack( padx=50, pady=10 )

        texts = [ "Save Data", "Save Figure" ]
        files = [ "UV_data.txt", "UV_figure.png" ]
        out_dir = get_setting( 'analysis_folder' )
        cb_vars = []
        file_vars = []
        row = 0
        for i, t in enumerate(texts):
            cb_var = Tk.IntVar()
            cb_vars.append( cb_var )
            cb = Tk.Checkbutton( detail_frame, text=t,
                                    variable=cb_var )
            cb.grid( row=row, column=0, sticky=Tk.W )
            file = Tk.StringVar()
            path = os.path.join( out_dir , files[i] ).replace( '\\', '/' )
            file.set( path )
            file_vars.append( file )
            file_entry = FileEntry( detail_frame, textvariable=file, width=60 )
            file_entry.grid( row=row, column=1, sticky=Tk.W, padx=10 )
            row += 1
            if i == 0:
                restrict_frame = Tk.Frame( detail_frame )
                restrict_frame.grid( row=row, column=1 )
                row += 1

                restrict_var = Tk.IntVar()
                cb = Tk.Checkbutton( restrict_frame, text="restrict data so that wavelength(λ) is within", variable=restrict_var )
                cb.grid( row=0, column=0, sticky=Tk.W )

                range_vars = []
                col = 1
                for k in range(2):
                    var = Tk.DoubleVar()
                    var.set( int( self.parent.wl_vector[ 0  if k==0 else -1] ) )
                    range_vars.append( var )
                    entry = Tk.Entry( restrict_frame, textvariable=var, width=6, justify=Tk.CENTER )
                    entry.grid( row=0, column=col, padx=5 )
                    if k == 0:
                        col += 1
                        label = Tk.Label( restrict_frame, text="and" )
                        label.grid( row=0, column=col )
                    col += 1

                space = Tk.Frame( detail_frame, height=10 )
                space.grid( row=row, column=0 )
                row += 1

        self.cb_vars    = cb_vars
        self.file_vars  = file_vars
        self.restrict_var   = restrict_var
        self.range_vars = range_vars

    def validate( self ):
        if self.restrict_var.get() == 1:
            f = self.range_vars[0].get()
            t = self.range_vars[1].get()
            range_ = np.where( self.parent.wl_vector >= f and self.parent.wl_vector <= t )[0]
            slice_ = slice(range_[0], range_[-1]+1)
        else:
            slice_ = slice(None, None)

        for k, cb_var in enumerate(self.cb_vars):
            if cb_var.get() == 0:
                continue

            file = self.file_vars[k].get()
            if k == 0:
                wl_vector = self.parent.wl_vector[slice_]
                wl_vector.reshape( ( 1, len(wl_vector )) )
                absorbance  = self.parent.absorbance
                corrected_data  = absorbance.get_corrected_data()
                col_header      = absorbance.col_header

                data    = corrected_data[slice_,:].T

                if False:
                    from mpl_toolkits.mplot3d   import Axes3D
                    import molass_legacy.KekLib.DebugPlot as dplt

                    dplt.push()
                    i = np.arange(data.shape[1])
                    y = np.arange(data.shape[0])
                    ii, yy = np.meshgrid(i, y)
                    zz = data[yy,ii]
                    w = self.parent.wl_vector[slice_]
                    xx, yy = np.meshgrid(w, y)
                    fig = dplt.figure()
                    ax = fig.add_subplot(111, projection='3d')
                    ax.plot_surface(xx, yy, zz, alpha=0.3)
                    fig.tight_layout()
                    dplt.show()
                    dplt.pop()

                data_   = np.vstack( [ wl_vector, data ] ).T
                col_header = np.array( [ '' ] + absorbance.col_header )
                # print( 'len(col_header)=', len(col_header) )
                # print( 'data_.shape=', data_.shape )

                with open(file, 'w', encoding='cp932') as fh:
                    fh.write('>>>>>>>>>>>>>> 連続計測（領域波長）Data Start<<<<<<<<<<<<\n')
                    delimter = ',' if file.lower().find( '.csv' ) > 0 else '\t'
                    fh.write( delimter.join(col_header) + '\n' )

                np_savetxt(file, data_, mode='a')

                with open(file, 'a', encoding='cp932') as fh:
                    fh.write('>>>>>>>>>>>>>> Data End <<<<<<<<<<<<\n')

            else:
                self.parent.fig.savefig( file )

        return 1



import sys
import os
import numpy    as np
from mpl_toolkits.mplot3d   import Axes3D, proj3d
import matplotlib.pyplot    as plt
# import statsmodels.api      as sm
from Intensity              import Intensity

sys.path.append( os.path.dirname( os.path.abspath( __file__ ) ) + '/../../../lib' )
import molass_legacy.KekLib
import OurStatsModels       as sm
from molass_legacy.KekLib.NumpyUtils             import np_savetxt

max_intensity = 0.03
error_level = 0.002
error_level_100 = error_level / np.sqrt(100)

# np.random.seed( 1234 )

def extrapolate( X, ye_array ):
    a_array = []
    for j in range(len(q)):
        y = ye_array[:,j]
        w = np.ones( (len(y),) ) * 1/error_level_100**2
        model   = sm.WLS(y, X, weights=w)
        result  = model.fit()
        a_array.append( result.params[0] )

    return np.array( a_array )

def plot_extrapolated( ax, y_, color='blue' ):
    ax.set_title( 'extrapolated intensity/c' )
    max_x_intensity = 1.1
    ax.set_ylim( 0, max_x_intensity )
    ax.plot( q, y_, color=color )
    ax.plot( [ q_see, q_see ], [ 0, max_x_intensity ], ':r' )
    ax.plot( q_see, y_[q_see_i], 'or' )

def annotate_error( ax, x, y, text, offset_ratios=[0.1, 0.1], ha='left' ):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    xoffset = ( xmax - xmin ) * offset_ratios[0]
    yoffset = ( ymax - ymin ) * offset_ratios[1]

    ax.annotate( text, xy=(x, y),
                    xytext=( x + xoffset, y + yoffset ),
                    ha=ha,
                    arrowprops=dict( headwidth=5, width=0.5, facecolor='black', shrink=0.05),
                    )

fig = plt.figure( figsize=(18, 9) )

axes = (    [ fig.add_subplot( 4, 5, i+1, projection='3d' ) for i in range(4) ] + [ None ]
        +   [ fig.add_subplot( 4, 5, i+6 ) for i in range(5) ]
        +   [ fig.add_subplot( 4, 5, i+11, projection='3d' ) for i in range(4) ] + [ None ]
        +   [ fig.add_subplot( 4, 5, i+16 ) for i in range(5) ]
       )

qsize = 0.3
xmin = -qsize
xmax =  qsize
ymin = -qsize
ymax =  qsize
N = 100
c = 0.01
R = 29

q = np.linspace(0, qsize, N)
q_see_i = 15
q_see = q[q_see_i]
xx, yy = np.meshgrid( np.linspace(xmin, xmax, N), np.linspace(ymin, ymax, N) )
s = np.sqrt( xx**2 + yy**2 )

tt = np.linspace( 0, 2*np.pi, 100 )
tx = q_see * np.cos( tt )
ty = q_see * np.sin( tt )
ts = np.sqrt( tx**2 + ty**2 )

cvector = np.linspace( 0.01, 0.04, 4 )

y0_array = []
ye_array = []

for i, c in enumerate( cvector ):
    intensity = Intensity( c, R )
    zz = intensity( s )
    ax1 = axes[i]
    ax1.set_title( 'model intensity for c=%.2g' % c )
    ax1.set_xlim( -qsize, qsize )
    ax1.set_ylim( -qsize, qsize )
    ax1.set_zlim( 0, max_intensity )
    ax1.plot_surface(xx, yy, zz, alpha=0.1 )
    tz = intensity(ts)
    ax1.plot( tx, ty, tz, color='red' )

    ax2 = axes[i+5]
    ax2.set_title( 'model intensity for c=%.2g' % c )
    ax2.set_xlim( 0, qsize )
    ax2.set_ylim( 0, max_intensity )
    y = intensity( q )
    y0_array.append( y )
    ax2.plot( q, y )
    ax2.plot( [ q_see, q_see ], [ 0, max_intensity ], ':r' )
    ax2.plot( q_see, y[q_see_i], 'or' )

    ax3 = axes[i+10]
    ax3.set_title( 'noisy intensity for c=%.2g' % c )
    ax3.set_xlim( -qsize, qsize )
    ax3.set_ylim( -qsize, qsize )
    ax3.set_zlim( 0, max_intensity )
    ee = error_level * np.random.randn( len(zz) )
    ax3.plot_surface(xx, yy, zz+ee, alpha=0.1 )
    te = error_level * np.random.randn( len(ts) )
    ax3.plot( tx, ty, tz+te, color='red' )

    tx_, ty_, _ = proj3d.proj_transform(tx[0],ty[0],tz[0], ax3.get_proj())
    annotate_error( ax3, tx_, ty_, 'error=%.3g' % ( error_level ),
                        offset_ratios=[0.02, 0.05], ha='center' )

    ax4 = axes[i+15]
    ax4.set_title( 'noisy intensity for c=%.2g' % c )
    ax4.set_xlim( 0, qsize )
    ax4.set_ylim( 0, max_intensity )
    y = intensity( q )
    e = error_level_100 * np.random.randn( len(q) )
    ye = y + e
    ye_array.append( ye )
    ax4.plot( q, y+e )
    ax4.plot( [ q_see, q_see ], [ 0, max_intensity ], ':r' )
    ax4.plot( q_see, ye[q_see_i], 'or' )
    annotate_error( ax4, q_see, ye[q_see_i],
                        r'error=$\frac{%.3g}{\sqrt{100}}$=%.3g' % ( error_level, error_level_100 ),
                        )

X = np.array( [ cvector, -cvector**2 ] ).T
y0_array_ = np.array(y0_array)
ye_array_ = np.array(ye_array)

y0_ = extrapolate( X, y0_array_ )
plot_extrapolated( axes[ 9], y0_ )

ye_ = extrapolate( X, ye_array_ )
plot_extrapolated( axes[19], ye_ )

y_ex_array = []
for i in range(1000):
    ye_array = []
    for c in cvector:
        intensity = Intensity( c, R )
        y = intensity( q )
        e = error_level_100 * np.random.randn( len(q) )
        ye_array.append( y + e )

    y_ = extrapolate( X, np.array(ye_array) )
    y_ex_array.append( y_ )

y_ex_array_ = np.array( y_ex_array )
y_ex_ = np.average( y_ex_array_, axis=0 )
e_ex_ = np.std( y_ex_array_, axis=0 )
ax_ex = axes[19]

ax_ex.errorbar( q, y_ex_, yerr=e_ex_, color='yellow',  fmt='none' )
plot_extrapolated( ax_ex, y_ex_, color='cyan' )
annotate_error( ax_ex, q_see, y_ex_[q_see_i], 'error=?=%.3g' % ( e_ex_[q_see_i] ) )

y = ye_array_[:,q_see_i]
N = len(y)
e2 = np.ones( (N,) ) * error_level_100**2
w = 1/e2

model   = sm.WLS(y, X, weights=w)
result  = model.fit()
param_error = np.sqrt( np.diag( result.cov_params() ) ) 
print( 'param error=', param_error )

# e2 = ( np.dot( X, result.params.T ) - y )**2
print( 'N=', N)
error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
print( 'error=', error )
print( 'ratio=', param_error/error )


plt.tight_layout()
plt.show()



"""
    SdUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

def get_sd(in_folder, trimming=True, correction=True):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl

    set_setting("in_folder", in_folder)

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd)
    if trimming:
        sd_ = sd._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd.get_copy()

    if correction:
        v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
    else:
        v2_copy = sd_

    return v2_copy



"""
    Optimizer.PenaltyUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy.Optimizer.XrUvScaleRatio import xruv_scale_ratio_penalty
PENALTY_SCALE = 1e3
UV_B_ALLOW_RATIO = 0.1

def compute_mapping_penalty(uv_curve, xr_curve, init_mapping, mapping, uv_x_size, xr_scales, uv_scales, debug=False):
    a_init, b_init = init_mapping
    a, b = mapping
    ratio = a/a_init
    a_deviation = min(0, ratio - 0.99)**2 + max(0, ratio - 1.05)**2
    b_allowance = uv_x_size * UV_B_ALLOW_RATIO
    b_deviation = max(0, abs(b - b_init) - b_allowance)**2

    ratio_penalty = xruv_scale_ratio_penalty(xr_scales, uv_scales, debug=debug)
    mapping_penalty = PENALTY_SCALE * (a_deviation + b_deviation ) + ratio_penalty

    if debug:
        print("ratio=", ratio)
        print("a_deviation=", a_deviation)
        print("b_allowance=", b_allowance)
        print("ratio_penalty=", ratio_penalty)
        print("mapping_penalty=", mapping_penalty)
        xr_x = xr_curve.x
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("compute_mapping_penalty: debug")
            ax.set_xlabel("XR")
            ax.set_ylabel("UV")
            ax.plot(xr_x, xr_x*a_init + b_init, label='initial mapping')
            ax.plot(xr_x, xr_x*a + b, ':', label='current mapping')
            fig.tight_layout()
            plt.show()

    return mapping_penalty


"""
    Alsaker2017_Test.py

    Copyright (c) 2024, SAXS Team, KEK-PF

    How to test

    R version:
        on VS-code R-debugger in the TODO/Rg code
        setwd('D:\\TODO\\20240523\\Rg code')
        data = read.table("oval_01C_S008_0_01.dat", header = FALSE)
        source('file1.R')
        estimate_Rg(data, 1, 5)

    Python version:
        on Command Pronmpt in the test folder
        python .../GuinierTools/Alsaker2017_Test.py
"""
import numpy as np

def load_data(file):
    data = np.loadtxt(file)
    return data

def demo_impl():
    from molass_legacy.SerialAnalyzer.DataUtils import get_local_path
    from importlib import reload
    import GuinierTools.Alsaker2017 as module
    reload(module)
    from GuinierTools.Alsaker2017 import estimate_Rg
    TODO = get_local_path('TODO')
    data = load_data(TODO + r'\20240523\Rg code\oval_01C_S008_0_01.dat')
    print(data.shape)
    estimate_Rg(data, 1, 5)

def demo():
    import molass_legacy.KekLib.DebugPlot as dpl

    extra_button_specs = [("Demo", demo_impl)]
    with dpl.Dp(extra_button_specs=extra_button_specs):
        fig, ax = dpl.subplots()
        dpl.show()

if __name__ == '__main__':
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")
    demo()


"""
    Peaks.EditCanvas.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, ToolTip
from molass_legacy._MOLASS.SerialSettings import get_setting
import molass_legacy.KekLib.CustomMessageBox as MessageBox
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.Peaks.PeakParamsSet import PeakParamsSet

class EditCanvas(Dialog):
    def __init__(self, parent, sd, ecurves, original_info, working_info=None):
        self.parent = parent
        self.sd = sd
        self.ecurves = ecurves
        self.xr_peaks_orig = original_info[2]   # xr_peaks to reset to
        if working_info is None:
            working_info = original_info
        self.xr_peaks_work = working_info[2]    # xr_peaks to undo to 
        working_info[2] = self.xr_peaks_work.copy()
        self.xr_draw_info = working_info
        self.applied = False
        self.peak_params_set = None
        self.appliable = False
        Dialog.__init__(self, parent, "Edit Canvas", visible=False)

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig)
        # print("ExtrapolSolverDialog: closed fig")
        Dialog.cancel(self)

    def show( self ):
        self._show()

    def body(self, body_frame, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.KekLib.DraggableCurves
            reload(KekLib.DraggableCurves)
        from molass_legacy.KekLib.DraggableCurves import DraggableCurves

        upper_frame = Tk.Frame(body_frame)
        upper_frame.pack(fill=Tk.X)
        lower_frame = Tk.Frame(body_frame)
        lower_frame.pack()
        self.upper_frame = upper_frame

        cframe = Tk.Frame(upper_frame)
        cframe.pack(side=Tk.LEFT)
        cframe_ = Tk.Frame(cframe)
        cframe_.pack()
        tframe_ = Tk.Frame(cframe)
        tframe_.pack(side=Tk.LEFT, fill=Tk.X)

        fig, ax = plt.subplots(figsize=(12,8))
        self.fig = fig
        self.ax = ax
        self.axt = ax.twinx()
        # https://stackoverflow.com/questions/55565393/matplotlib-picker-event-on-secondary-y-axis
        ax.set_zorder(self.axt.get_zorder() + 1)

        self.dcurves = DraggableCurves(ax, x_only=True,
                            button_release_user_callback=self.button_release_user_callback,
                            debug=True)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe_)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.draw_components()

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe_)
        self.toolbar.update()

        dcurves = self.dcurves
        # self.mpl_canvas.mpl_connect('button_press_event', self.on_click)
        self.mpl_canvas.mpl_connect('button_release_event', dcurves.button_release_callback)
        self.mpl_canvas.mpl_connect('pick_event', dcurves.pick_callback)
        self.mpl_canvas.mpl_connect('motion_notify_event', dcurves.motion_notify_callback)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=20)

        num_buttons = 5
        for j in range(num_buttons):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="Cancel", width=12, command=self.user_cancel)
        col = 0
        w.grid(row=0, column=col, pady=10)
        self.cancel_btn = w
        ToolTip(w, "discard all changes here")

        col += 1
        w = Tk.Button(box, text="Reset", width=12, command=self.reset)
        w.grid(row=0, column=col, pady=10)
        self.reset_btn = w
        ToolTip(w, "reset to the initial state of Peak Editor")

        col += 1
        w = Tk.Button(box, text="Undo", width=12, command=self.undo)
        w.grid(row=0, column=col, pady=10)
        self.undo_btn = w
        ToolTip(w, "undo to the initial state here")

        col += 1
        w = Tk.Button(box, text="Adjust", width=10, command=self.adjust_heights)
        w.grid(row=0, column=col, pady=10)
        self.adjust_btn = w
        ToolTip(w, "optimize parameters while fixing peak top positions")

        col += 1
        w = Tk.Button(box, text="Ok", width=12, command=self.ok)
        w.grid(row=0, column=col, pady=10)
        self.proceed_btn = w
        ToolTip(w, "aplly the change result here")

        assert col+1 == num_buttons

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.user_cancel)
        self.bind("<Control-z>", lambda event: self.undo())

    def draw_components(self, devel=False):
        if devel:
            from importlib import reload
            import QuickAnalysis.ModeledPeaks
            reload(QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import plot_curve

        ax = self.ax
        axt = self.axt
        ax.cla()
        axt.cla()
        axt.grid(False)

        ax.set_title("Xray Elution at Q=%.3g" % get_setting("intensity_picking"), fontsize=16)

        self.dcurves.clear()
        xr_x, xr_y, xr_peaks, xr_baseline = self.xr_draw_info
        self.markers = plot_curve(ax, xr_x, xr_y, xr_peaks, color='orange', baseline=xr_baseline,
                    dcurves=self.dcurves, return_markers=True)
        peaktops = []
        for marker in self.markers:
            xdata, ydata = marker.get_data()
            peaktops.append((xdata[0], ydata[0]))
        self.peaktops = np.array(peaktops)

        ax.set_xlim(*ax.get_xlim())     # fix for editting
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def button_release_user_callback(self, dcurves):
        print("button_release_user_callback")
        dx, dy = dcurves.get_last_displacement()
        x, y = dcurves.picked_artist.get_data()
        x = x - dx  # note that x -= dx would change the artist's data
        j = np.argmax(y)
        px = x[j]
        py = y[j]
        diffs = self.peaktops - np.array((px,py))
        dists = diffs[:,0]**2 + diffs[:,1]**2
        m = np.argmin(dists)
        self.markers[m].set_data(px+dx, py)

    def user_cancel(self, ask=True):
        self.cancel()

    def reset(self):
        self.appliable = False
        self.xr_draw_info[2] = self.xr_peaks_orig
        self.draw_components()

    def undo(self):
        self.xr_draw_info[2] = self.xr_peaks_work
        self.draw_components()

    def adjust_heights(self, debug=False):
        if debug:
            from importlib import reload
            import QuickAnalysis.ModeledPeaks
            reload(QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import adjust_peak_heights, plot_curve

        uv_curve, xr_curve = self.ecurves
        xr_x = xr_curve.x
        xr_y = xr_curve.y
        uv_x = uv_curve.x
        uv_y = uv_curve.y

        self.update_xr_peaks()
        a, b = self.parent.get_pre_recog_mapping_params()
        new_peaks = self.get_xr_peaks()
        new_xr_peaks = adjust_peak_heights(1, 0, xr_x, xr_y, new_peaks, xr_x, xr_y, debug=debug)
        new_uv_peaks = adjust_peak_heights(a, b, xr_x, xr_y, new_xr_peaks, uv_x, uv_y, debug=debug)
        if debug:
            with dplt.Dp():
                fig, (ax1, ax2) = dplt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("adjust_heights debug")
                plot_curve(ax1, uv_x, uv_y, new_uv_peaks, color='blue')
                plot_curve(ax2, xr_x, xr_y, new_xr_peaks, color='orange')
                dplt.show()
        self.peak_params_set = PeakParamsSet(new_uv_peaks, new_xr_peaks, a, b)
        self.xr_peaks_work = new_xr_peaks
        # self.update_xr_peaks()
        self.xr_draw_info[2] = self.xr_peaks_work
        self.draw_components()
        self.appliable = True

    def get_peak_params_set(self):
        return self.peak_params_set

    def validate(self):
        if self.appliable:
            return 1
        else:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            MessageBox.showinfo("Notification",
            'Please press "Adjust" button to confirm the edit result\n'
            'before applying.',
            parent=self,
            )
            return 0

    def apply(self, show_message=False):
        self.applied = True
        if show_message:
            MessageBox.showinfo("Remainder",
                "Remember that the results will be sorted, locally optimized, and renumbered.",
                parent=self)
        self.update_xr_peaks()

    def update_xr_peaks(self):
        displ_list = self.dcurves.get_displacements()
        print("displacements: ", displ_list)
        new_xr_peaks = []
        for k, params in enumerate(self.xr_peaks_work):
            temp_params = np.array(params)
            # note that temp_params[1] for both affine and non-affine models
            temp_params[1] += displ_list[k][0]      # move in x-coordinate only
            new_xr_peaks.append(temp_params)
        self.xr_peaks_work = np.array(sorted(new_xr_peaks, key=lambda p: p[1]))

    def get_xr_peaks(self):
        return self.xr_peaks_work



"""
    Models.Stochastic.TheoryTutorial.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

def tutorial_illust():
    from molass_legacy.Models.Stochastic.Monopore import monopore_func
    print("tutorial_illust")

    N = 1000
    T = 0.5
    t0 = 50
    me = 1.5
    mp = 1.5
    poresize = 100

    x = np.arange(600)
    cy_list = []
    rgs = [44, 40, 36, 20]
    scales = [0.1, 0.5, 0.2, 0.1]
    for (rg, scale) in zip(rgs,scales):
        cy = monopore_func(x, scale, N, T, t0, me, mp, poresize, rg)
        cy_list.append(cy)
    y = np.sum(cy_list, axis=0)

    def compute_monopore_tv(rv):
        rhov = np.asarray(rv)/poresize
        rhov[rhov > 1] = 1
        return t0 + N * T * (1 - rhov)**(me + mp)

    rv = np.linspace(1, 75, 100)
    tv = compute_monopore_tv(rv)
    pv = compute_monopore_tv(rgs)

    fig, ax = plt.subplots(figsize=(10,5))
    ax.set_title("Stochastic Theory Illustration where N=%g, T=%g, $t_0$=%g, $m_e$=%g, $m_p$=%g, poresize=%g"
                 % (N, T, t0, me, mp, poresize), y=1.05, fontsize=14)
    ax.set_xlabel("Time (Frames)")
    ax.set_ylabel("Density")
    for k, (rg, cy) in enumerate(zip(rgs, cy_list), start=1):
        ax.plot(x, cy, ":", lw=2, label="component-%d ($R_g$=%g)" %(k, rg))

    ax.plot(x, y, color="orange", alpha=0.5, label="component total")
    ax.axvline(x=t0, color="red", label="$t_0$ = %g" % t0, alpha=0.5)
    ax.axvline(x=t0 + N*T, color="green", label="$t_0$ + N*T = %g" % (t0+N*T), alpha=0.5)
    ax.legend(loc="upper center")

    axt = ax.twinx()
    axt.grid(False)
    axt.set_ylabel(r"$R_g (\AA_)$")

    axt.plot(tv, rv, color="yellow", label="Exclusion Curve")
    axt.plot(pv, rgs, "o", markersize=3, color="red", label="Component Peak Positions")
    dx, dy = 5, 2
    for tx, ty in zip(pv, rgs):
        axt.text(tx+dx, ty+dy, "%g" % ty, alpha=0.5)
    axt.legend()

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin, 0.06)
    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    import seaborn as sns
    sns.set_theme()
    sys.path.append("../lib")

    tutorial_illust()


# coding: utf-8
"""
    DmaxDependency.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.SerialSettings import set_setting
from DataUtils import serial_folder_walk
from molass_legacy.Batch.StandardProcedure import StandardProcedure

sp = StandardProcedure()

dmax_variations = [80, 90, 100, 110, 120]

def run_datgnom(result):
    print('run_datgnom: result=', result)

def run_denss_for_survey(k, result, dmax):
    print(dmax)

num_folders = 0

def run_denss_varying_dmax(folder, *args, **kwargv):
    global num_folders
    num_folders += 1
    print([num_folders], folder)
    set_setting('in_folder', folder)

    sp.load(folder)
    sd = sp.get_corrected_sd()
    preview_results = sp.get_preview_results(sd)
    for result in preview_results:
        run_datgnom(result)
        continue

        for dmax in dmax_variations:
            for k in range(2):
                run_denss_for_survey(k, preview_results, dmax)

    return False, None

def collect_dependecy_data(root_folder):
    serial_folder_walk(root_folder, run_denss_varying_dmax)



"""
    Optimizer.FvScoreDetails.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import re
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from tksheet import Sheet
from .BasicOptimizer import PARAMS_SCALE

class FvScoreDetails(Dialog):
    def __init__(self, parent, optimizer, params):
        self.parent = parent
        self.optimizer = optimizer
        self.params = params
        self.separate_params = optimizer.split_params_simple(params)
        self.lower_bounds = optimizer.lower_bounds
        self.upper_bounds = optimizer.upper_bounds

        Dialog.__init__(self, parent, "FV Score Details", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        fv, score_list, Pxr, Cxr, Puv, Cuv, mapped_UvD = self.optimizer.objective_func(self.params, return_full=True)
        score_names = self.optimizer.get_score_names()

        data_list = []
        data_list.append(["", "Score Name", "Score Value"])
        for name, value in zip(score_names, score_list):
            row = ["", name, "%g" % value]
            if name == "Guinier_deviation":
                dev1, dev2 = self.optimizer.gdev.compute_deviation(Pxr, Cxr, self.separate_params[2], return_details=True)
                row += ["", "%g" % dev1, "%g" % dev2]
            data_list.append(row)

        data_list[1][0] = "Score Details"

        data_list.append([])

        data_list.append(["FV", "%g" % fv])

        data_list.append([])
        data_list.append(["", "Parameter Name", "Real Value", "Normalized Value", "Bounds Mask", "Lower Bound", "Upper Bound", "Penalty Type"])
        start_row_params = len(data_list)

        delchars_re = re.compile(r"[\$\\_{}]")
        param_names = [re.sub(delchars_re, "", name) for name in self.optimizer.get_parameter_names()]
        norm_values = self.optimizer.to_norm_params(self.params)
        mask_index = 0
        for name, real_value, norm_value, bnd_mask in zip(param_names, self.params, norm_values, self.optimizer.bounds_mask):
            if bnd_mask:
                lower_bound = self.lower_bounds[mask_index]
                upper_bound = self.upper_bounds[mask_index]
                mark = "" if real_value >= lower_bound and real_value <= upper_bound  else "out_of_bounds"
                mask_index += 1
                lower_bound_ = "%g" % lower_bound
                upper_bound_ = "%g" % upper_bound
            else:
                mark = "negative" if (name not in ["x0", "t0", "xba", "xbb", "s1", "s2", "mpb", "", "mra", "mrb", "diffratio", "Ti"] and real_value < 0) else ""
                lower_bound_ = ""
                upper_bound_ = ""
            data_list.append(["", name, "%g" % real_value, "%g" % norm_value, str(bnd_mask), lower_bound_, upper_bound_, mark])

        data_list[start_row_params][0] = "Parameter Details"

        num_columns = 8
        num_valid_rows = min(30, len(score_list) + 1 + len(param_names))

        column_width = 150
        width = column_width*num_columns + 60
        height = int(22*num_valid_rows) + 60
        self.sheet = Sheet(body_frame, width=width, height=height, data=data_list, show_selected_cells_border=False, column_width=column_width)
        self.sheet.pack()
        self.sheet.enable_bindings()



"""
    Models.Stochastic.PairedAnalysisSpecs.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
PORESIZE_BOUNDS = 75, 100
NUM_PLATES = 20000      # 14400*1.36 = 19584, statinary 0.6**2 : mobile 1.0**2
NUM_PLATES_THIN = 5000

def get_exec_spec(n):
    from molass_legacy.SerialAnalyzer.DataUtils import get_local_path

    PyTools = get_local_path('PyTools')

    exec_spec_list = [
        {
        "input_folders" :       [ PyTools + r"\Data\20210727\data01",   PyTools + r"\Data\20210727\data02"],
        "num_components" :      4,
        "unreliable_indeces" :  [1, 2],
        "poresize_bounds" :     PORESIZE_BOUNDS,
        "init_N0" :             [NUM_PLATES, NUM_PLATES_THIN ],
        },

        {
        "input_folders" :       [ PyTools + r"\Data\20220716\BSA_201",   PyTools + r"\Data\20220716\BSA_202"],
        "num_components" :      4,
        "unreliable_indeces" :  [0],
        "poresize_bounds" :     PORESIZE_BOUNDS,
        "init_N0" :             [NUM_PLATES, NUM_PLATES_THIN],
        },

        {
        "input_folders" :       [ PyTools + r"\Data\20220716\FER_OA_301",   PyTools + r"\Data\20220716\FER_OA_302"],
        "num_components" :      4,
        "unreliable_indeces" :  [0, 2],
        "poresize_bounds" :     PORESIZE_BOUNDS,
        "init_N0" :             [NUM_PLATES, NUM_PLATES_THIN],
        },

        {
        "input_folders" :       [ PyTools + r"\Data\20220716\OA_ALD_201",   PyTools + r"\Data\20220716\OA_ALD_202"],
        "num_components" :      4,
        "unreliable_indeces" :  [0, 2],
        "use_mapping" : True,
        "poresize_bounds" :     PORESIZE_BOUNDS,
        "init_N0" :             [NUM_PLATES, NUM_PLATES_THIN],
        },

        {
        "input_folders" :       [ PyTools + r"\Data\20230706\ALD_OA001",   PyTools + r"\Data\20230706\ALD_OA002"],
        "num_components" :      3,
        "unreliable_indeces" :  [],
        "poresize_bounds" :     (75, 300),
        "init_N0" :             [20000, 20000],
        },

        {
        "input_folders" :       [ PyTools + r"\Data\20230706\BSA001",   PyTools + r"\Data\20230706\BSA002"],
        "num_components" :      4,
        "components_to_use" :   [[0, 1, 2, 3], None],
        "unreliable_indeces" :  [0],
        "poresize_bounds" :     (75, 300),
        "init_N0" :             [20000, 20000],
        },
    ]

    return exec_spec_list[n]

def get_components_to_use(exec_spec):
    components_to_use = exec_spec.get('components_to_use')
    if components_to_use is None:
        return [None, None]
    else:
        return components_to_use